# coding: utf-8

"""
Code to analyse VASP defect calculations.
These functions are built from a combination of useful modules from pymatgen, alongside
substantial modification, in the efforts of making an efficient, user-friendly package for
managing and analysing defect calculations, with publication-quality outputs
"""

import copy
import os
import warnings
from operator import itemgetter

import pandas as pd
import numpy as np
from pymatgen.analysis.defects.thermodynamics import DefectPhaseDiagram
from pymatgen.analysis.defects.utils import TopographyAnalyzer
from pymatgen.util.string import unicodeify
from pymatgen.io.vasp.inputs import Poscar
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.analysis.defects.core import DefectEntry
from monty.serialization import loadfn, dumpfn
from monty.json import MontyDecoder
from tabulate import tabulate

from doped.pycdt.utils.vasp import DefectRelaxSet
from doped.pycdt.utils import parse_calculations
from doped import _ignore_pmg_warnings

_ANGSTROM = "\u212B"  # unicode symbol for angstrom to print in strings
_ignore_pmg_warnings()  # ignore unnecessary pymatgen warnings


def bold_print(string: str) -> None:
    """Does what it says on the tin. Prints the input string in bold."""
    print("\033[1m" + string + "\033[0m")


def _convert_dielectric_to_tensor(dielectric):
    # check if dielectric in required 3x3 matrix format
    if not isinstance(dielectric, (float, int)):
        dielectric = np.array(dielectric)
        if dielectric.shape == (3,):
            dielectric = np.diag(dielectric)
        elif dielectric.shape != (3, 3):
            raise ValueError(
                f"Dielectric constant must be a float/int or a 3x1 matrix or 3x3 matrix, "
                f"got type {type(dielectric)} and shape {dielectric.shape}"
            )

    else:
        dielectric = np.eye(3) * dielectric

    return dielectric


# TODO: Should add check the bulk and defect KPOINTS/INCAR/POTCAR/POSCAR (size) settings
#  are compatible, and throw warning if not.
# TODO: Add `check_defects_compatibility()` function that checks the bulk and defect
#  KPOINTS/INCAR/POTCAR/POSCAR (size) settings for all defects in the supplied defect_dict
#  are compatible, if not throw warnings and say what the differences are. Should recommend
#  using this in the example notebook if a user has parsed the defects individually (rather
#  than with the single looping function described below):
# TODO: Add a function that loops over all the defects in a directory (with `defect_dir = .`,
#  and `subfolder = vasp_ncl` options) and parses them all (use `SnB` defect-name-matching
#  function?), returning a dictionary of defect entries, with the defect name as the key. (i.e.
#  doing the loop in the example notebook). Show both this function and the individual function
#  calls in the example notebook. Benefit of this one is that we can then auto-run
#  `check_defects_compatibility()` at the end of parsing the full defects dict. – When doing
#  this, look at `PostProcess` code, and then delete it once all functionality is moved here.
def defect_entry_from_paths(
    defect_path,
    bulk_path,
    dielectric,
    charge=None,
    initial_defect_structure=None,
    skip_corrections=False,
    bulk_bandgap_path=None,
    **kwargs,
):
    """
    Parse the defect calculation outputs in `defect_path` and return the parsed `DefectEntry`
    object.

    Args:
    defect_path (str): path to defect folder of interest (with vasprun.xml(.gz))
    bulk_path (str): path to bulk folder of interest (with vasprun.xml(.gz))
    dielectric (float or int or 3x1 matrix or 3x3 matrix):
        ionic + static contributions to dielectric constant
    charge (int): charge of defect. If not provided, will be automatically determined
        from the defect calculation outputs (requires POTCARs to be set up with `pymatgen`).
    initial_defect_structure (str):  Path to the unrelaxed defect structure,
        if structure matching with the relaxed defect structure(s) fails.
    skip_corrections (bool): Whether to skip the calculation and application of finite-size
        charge corrections to the defect energy.
    bulk_bandgap_path (str): Path to bulk OUTCAR file for determining the band gap. If the
        VBM/CBM occur at reciprocal space points not included in the bulk supercell calculation,
        you should use this tag to point to a bulk bandstructure calculation instead. If None,
        will use self.defect_entry.parameters["bulk_path"].

    Return:
        Parsed `DefectEntry` object.
    """
    _ignore_pmg_warnings()  # ignore unnecessary pymatgen warnings
    dielectric = _convert_dielectric_to_tensor(dielectric)

    parameters = {
        "bulk_path": bulk_path,
        "defect_path": defect_path,
        "dielectric": dielectric,
    }

    # add bulk simple properties
    bulk_vr_path, multiple = parse_calculations._get_output_files_and_check_if_multiple(
        "vasprun.xml", bulk_path
    )
    if multiple:
        warnings.warn(
            f"Multiple `vasprun.xml` files found in bulk directory: {bulk_path}. Using "
            f"{bulk_vr_path} to parse the calculation energy and metadata."
        )
    bulk_vr = parse_calculations.get_vasprun(bulk_vr_path)
    bulk_energy = bulk_vr.final_energy
    bulk_sc_structure = bulk_vr.initial_structure.copy()

    # add defect simple properties
    (
        defect_vr_path,
        multiple,
    ) = parse_calculations._get_output_files_and_check_if_multiple(
        "vasprun.xml", defect_path
    )
    if multiple:
        warnings.warn(
            f"Multiple `vasprun.xml` files found in defect directory: {defect_path}. Using"
            f" {defect_vr_path} to parse the calculation energy and metadata."
        )
    defect_vr = parse_calculations.get_vasprun(defect_vr_path)
    defect_energy = defect_vr.final_energy

    # get defect charge
    try:
        defect_nelect = defect_vr.incar.get("NELECT", None)
        if defect_nelect is None:
            auto_charge = 0  # neutral defect if NELECT not specified
        else:
            potcar_symbols = [titel.split()[1] for titel in defect_vr.potcar_symbols]
            potcar_settings = {
                symbol.split("_")[0]: symbol for symbol in potcar_symbols
            }
            neutral_defect_relax_set = DefectRelaxSet(
                defect_vr.structures[-1],
                charge=0,
                user_potcar_settings=potcar_settings,
            )
            try:
                auto_charge = -1 * (defect_nelect - neutral_defect_relax_set.nelect)

            except Exception as e:
                raise RuntimeError(
                    f"Defect charge cannot be automatically determined as POTCARs have not "
                    f"been setup with pymatgen (see Step 2 at "
                    f"https://github.com/SMTG-UCL/doped#installation). Please specify defect "
                    f"charge manually using the `charge` argument, or set up POTCARs with "
                    f"pymatgen. Got error: {e}"
                )

            if abs(auto_charge) >= 10:  # crazy charge state predicted
                raise RuntimeError(
                    f"Auto-determined defect charge q={int(auto_charge):+} is unreasonably large. "
                    f"Please specify defect charge manually using the `charge` argument."
                )

        if (
            charge is not None
            and int(charge) != int(auto_charge)
            and abs(auto_charge) < 5
        ):
            warnings.warn(
                f"Auto-determined defect charge q={int(auto_charge):+} does not match "
                f"specified charge q={int(charge):+}. Will continue with specified "
                f"charge, but beware!"
            )
        else:
            charge = auto_charge

    except Exception as e:
        if charge is None:
            raise e

    # Can specify initial defect structure (to help PyCDT find the defect site if
    # multiple relaxations were required, else use from defect relaxation OUTCAR:
    if initial_defect_structure:
        initial_defect_structure = Poscar.from_file(
            initial_defect_structure
        ).structure.copy()
    else:
        initial_defect_structure = defect_vr.initial_structure.copy()

    # Add initial defect structure to parameters, so it can be pulled later on
    # (eg. for Kumagai loader)
    parameters["initial_defect_structure"] = initial_defect_structure

    # identify defect site, structural information, and create defect object
    try:
        def_type, comp_diff = parse_calculations.get_defect_type_and_composition_diff(
            bulk_sc_structure, initial_defect_structure
        )
    except RuntimeError as exc:
        raise ValueError(
            "Could not identify defect type from number of sites in structure: "
            f"{len(bulk_sc_structure)} in bulk vs. {len(initial_defect_structure)} in defect?"
        ) from exc

    bulk_site_idx = None
    defect_site_idx = None
    unrelaxed_defect_structure = None
    # Try automatic defect site detection - this gives us the "unrelaxed" defect structure
    try:
        (
            bulk_site_idx,
            defect_site_idx,
            unrelaxed_defect_structure,
        ) = parse_calculations.get_defect_site_idxs_and_unrelaxed_structure(
            bulk_sc_structure, initial_defect_structure, def_type, comp_diff
        )
    except RuntimeError as exc:
        # if auto site-matching failed, try use transformation.json
        # The goal is to find the `defect_site_idx` or `defect_site_idx` based on the
        # tranformation.
        transformation_path = os.path.join(defect_path, "transformation.json")
        if not os.path.exists(transformation_path):  # try next folder up
            orig_transformation_path = transformation_path
            transformation_path = os.path.join(
                os.path.dirname(os.path.normpath(defect_path)),
                "transformation.json",
            )
            if os.path.exists(transformation_path):
                print(
                    f"No transformation file found at {orig_transformation_path}, but found "
                    f"one at {transformation_path}. Using this for defect parsing."
                )

        if os.path.exists(transformation_path):
            tf = loadfn(transformation_path)
            site = tf["defect_supercell_site"]
            if def_type == "vacancy":
                poss_deflist = sorted(
                    bulk_sc_structure.get_sites_in_sphere(
                        site.coords, 0.2, include_index=True
                    ),
                    key=lambda x: x[1],
                )
                searched = "bulk_supercell"
                if poss_deflist:
                    bulk_site_idx = poss_deflist[0][2]
            else:
                poss_deflist = sorted(
                    initial_defect_structure.get_sites_in_sphere(
                        site.coords, 2.5, include_index=True
                    ),
                    key=lambda x: x[1],
                )
                searched = "initial_defect_structure"
                if poss_deflist:
                    defect_site_idx = poss_deflist[0][2]
            if not poss_deflist:
                raise ValueError(
                    f"{transformation_path} specified defect site {site}, but could not find "
                    f"it in {searched}. Abandoning parsing."
                ) from exc
            if poss_deflist[0][1] > 1:
                site_matched_defect = poss_deflist[0]  # pymatgen Neighbor object
                offsite_warning = (
                    f"Site-matching has determined {site_matched_defect.species} at "
                    f"{site_matched_defect.coords} as the defect site, located "
                    f"{site_matched_defect.nn_distance:.2f} {_ANGSTROM} from "
                    f"its initial position. This may incur small errors in the charge correction."
                )
                warnings.warn(message=offsite_warning)

        else:
            raise RuntimeError(
                f"Could not identify {def_type} defect site in defect structure. "
                f"Try supplying the initial defect structure to "
                f"SingleDefectParser.from_paths(), or making sure the doped "
                f"transformation.json files are in the defect directory."
            ) from exc

    if def_type == "vacancy":
        defect_site = bulk_sc_structure[bulk_site_idx]
    else:
        if unrelaxed_defect_structure:
            defect_site = unrelaxed_defect_structure[defect_site_idx]
        else:
            defect_site = initial_defect_structure[defect_site_idx]

    if unrelaxed_defect_structure:
        if def_type == "interstitial":
            # get closest Voronoi site in bulk supercell to final interstitial site as this is
            # likely to be the initial interstitial site
            try:
                struc_and_node_dict = loadfn("./bulk_voronoi_nodes.json")
                if not StructureMatcher(
                    stol=0.05,
                    primitive_cell=False,
                    scale=False,
                    attempt_supercell=False,
                    allow_subset=False,
                ).fit(struc_and_node_dict["bulk_supercell"], bulk_sc_structure):
                    warnings.warn(
                        "Previous bulk_voronoi_nodes.json detected, but does not "
                        "match current bulk supercell. Recalculating Voronoi nodes."
                    )
                    raise FileNotFoundError

                voronoi_frac_coords = struc_and_node_dict["Voronoi nodes"]

            except FileNotFoundError:  # first time parsing
                topography = TopographyAnalyzer(
                    bulk_sc_structure,
                    bulk_sc_structure.symbol_set,
                    [],
                    check_volume=False,
                )
                topography.cluster_nodes()
                topography.remove_collisions()
                voronoi_frac_coords = [site.frac_coords for site in topography.vnodes]
                struc_and_node_dict = {
                    "bulk_supercell": bulk_sc_structure,
                    "Voronoi nodes": voronoi_frac_coords,
                }
                dumpfn(
                    struc_and_node_dict, "./bulk_voronoi_nodes.json"
                )  # for efficient
                # parsing of multiple defects at once
                print(
                    "Saving parsed Voronoi sites (for interstitial site-matching) to "
                    "bulk_voronoi_sites.json to speed up future parsing."
                )

            closest_node_frac_coords = min(
                voronoi_frac_coords,
                key=lambda node: defect_site.distance_and_image_from_frac_coords(node)[
                    0
                ],
            )
            int_site = unrelaxed_defect_structure[defect_site_idx]
            unrelaxed_defect_structure.remove_sites([defect_site_idx])
            unrelaxed_defect_structure.insert(
                defect_site_idx,  # Place defect at same position as in DFT calculation
                int_site.species_string,
                closest_node_frac_coords,
                coords_are_cartesian=False,
                validate_proximity=True,
            )
            defect_site = unrelaxed_defect_structure[defect_site_idx]

        # Use the unrelaxed_defect_structure to fix the initial defect structure
        initial_defect_structure = parse_calculations.reorder_unrelaxed_structure(
            unrelaxed_defect_structure, initial_defect_structure
        )
        parameters["initial_defect_structure"] = initial_defect_structure
        parameters["unrelaxed_defect_structure"] = unrelaxed_defect_structure
    else:
        warnings.warn(
            "Cannot determine the unrelaxed `initial_defect_structure`. Please ensure the "
            "`initial_defect_structure` is indeed unrelaxed."
        )

    for_monty_defect = {
        "@module": "pymatgen.analysis.defects.core",
        "@class": def_type.capitalize(),
        "charge": charge,
        "structure": bulk_sc_structure,
        "defect_site": defect_site,
    }
    defect = MontyDecoder().process_decoded(for_monty_defect)

    if unrelaxed_defect_structure:
        # only do StructureMatcher test if unrelaxed structure exists
        test_defect_structure = defect.generate_defect_structure()
        if not StructureMatcher(
            stol=0.25,
            primitive_cell=False,
            scale=False,
            attempt_supercell=False,
            allow_subset=False,
        ).fit(test_defect_structure, unrelaxed_defect_structure):
            # NOTE: this does not insure that cartesian coordinates or indexing are identical
            raise ValueError(
                "Error in defect object matching! Unrelaxed structure (1st below) "
                "does not match pymatgen defect.generate_defect_structure() "
                f"(2nd below):\n{unrelaxed_defect_structure}"
                f"\n{test_defect_structure}"
            )

    defect_entry = DefectEntry(
        defect, defect_energy - bulk_energy, corrections={}, parameters=parameters
    )

    return_sdp = kwargs.pop("return SingleDefectParser", False)  # internal use for tests/debugging
    sdp = parse_calculations.SingleDefectParser(
        defect_entry,
        defect_vr=defect_vr,
        bulk_vr=bulk_vr,
        **kwargs,  # in case user wants to specify `DefectCompatibility()`
    )
    if return_sdp:
        return sdp

    sdp.get_stdrd_metadata()  # Load standard defect metadata
    sdp.get_bulk_gap_data(bulk_bandgap_path=bulk_bandgap_path)  # Load band gap data

    if not skip_corrections:
        # determine charge correction to use, based on what output files are available (`LOCPOT`s or
        # `OUTCAR`s), and whether the supplied dielectric is isotropic or not
        def _check_folder_for_file_match(folder, filename):
            return any(
                filename.lower() in folder_filename.lower()
                for folder_filename in os.listdir(folder)
            )

        def _convert_anisotropic_dielectric_to_isotropic_harmonic_mean(
            aniso_dielectric,
        ):
            return 3 / sum(
                1 / diagonal_elt for diagonal_elt in np.diag(aniso_dielectric)
            )

        # check if dielectric (3x3 matrix) has diagonal elements that differ by more than 20%
        isotropic_dielectric = all(
            np.isclose(i, dielectric[0, 0], rtol=0.2) for i in np.diag(dielectric)
        )

        # regardless, try parsing OUTCAR files first (quickest)
        if _check_folder_for_file_match(
            defect_path, "OUTCAR"
        ) and _check_folder_for_file_match(bulk_path, "OUTCAR"):
            try:
                sdp.kumagai_loader()
            except Exception as kumagai_exc:
                if _check_folder_for_file_match(
                    defect_path, "LOCPOT"
                ) and _check_folder_for_file_match(bulk_path, "LOCPOT"):
                    try:
                        if not isotropic_dielectric:
                            # convert anisotropic dielectric to harmonic mean of the diagonal:
                            # (this is a better approximation than the pymatgen default of the
                            # standard arithmetic mean of the diagonal)
                            sdp.defect_entry.parameters[
                                "dielectric"
                            ] = _convert_anisotropic_dielectric_to_isotropic_harmonic_mean(
                                dielectric
                            )
                        sdp.freysoldt_loader()
                        if not isotropic_dielectric:
                            warnings.warn(
                                f"An anisotropic dielectric constant was supplied, but `OUTCAR` "
                                f"files (needed to compute the _anisotropic_ Kumagai eFNV charge "
                                f"correction) in the defect (at {defect_path}) & bulk (at "
                                f"{bulk_path}) folders were unable to be parsed, giving the "
                                f"following error message:\n{kumagai_exc}\n"
                                f"`LOCPOT` files were found in both defect & bulk folders, "
                                f"and so the Freysoldt (FNV) charge correction developed for "
                                f"_isotropic_ materials will be applied here, which corresponds "
                                f"to using the effective isotropic average of the supplied "
                                f"anisotropic dielectric. This could lead to significant errors "
                                f"for very anisotropic systems and/or relatively small supercells!"
                            )
                    except Exception as freysoldt_exc:
                        warnings.warn(
                            f"Got this error message when attempting to parse defect & bulk "
                            f"`OUTCAR` files to compute the Kumagai (eFNV) charge correction:"
                            f"\n{kumagai_exc}\nThen got this error message when attempting to "
                            f"parse defect & bulk `LOCPOT` files to compute the Freysoldt (FNV) "
                            f"charge correction:\n{freysoldt_exc}\n-> Charge corrections will not "
                            f"be applied for this defect."
                        )
                        if not isotropic_dielectric:
                            # reset dielectric to original anisotropic value if FNV failed as well:
                            sdp.defect_entry.parameters["dielectric"] = dielectric
                        skip_corrections = True

                else:
                    warnings.warn(
                        f"An anisotropic dielectric constant was supplied, but `OUTCAR` "
                        f"files (needed to compute the _anisotropic_ Kumagai eFNV charge "
                        f"correction) in the defect (at {defect_path}) & bulk (at "
                        f"{bulk_path}) folders were unable to be parsed, giving the "
                        f"following error message:\n{kumagai_exc}\n-> Charge corrections will not "
                        f"be applied for this defect."
                    )
                    skip_corrections = True

        elif _check_folder_for_file_match(
            defect_path, "LOCPOT"
        ) and _check_folder_for_file_match(bulk_path, "LOCPOT"):
            try:
                if not isotropic_dielectric:
                    # convert anisotropic dielectric to harmonic mean of the diagonal:
                    # (this is a better approximation than the pymatgen default of the
                    # standard arithmetic mean of the diagonal)
                    sdp.defect_entry.parameters[
                        "dielectric"
                    ] = _convert_anisotropic_dielectric_to_isotropic_harmonic_mean(
                        dielectric
                    )
                sdp.freysoldt_loader()
                if not isotropic_dielectric:
                    warnings.warn(
                        f"An anisotropic dielectric constant was supplied, but `OUTCAR` "
                        f"files (needed to compute the _anisotropic_ Kumagai eFNV charge "
                        f"correction) were not found in the defect (at {defect_path}) & bulk "
                        f"(at {bulk_path}) folders.\n"
                        f"`LOCPOT` files were found in both defect & bulk folders, "
                        f"and so the Freysoldt (FNV) charge correction developed for "
                        f"_isotropic_ materials will be applied here, which corresponds "
                        f"to using the effective isotropic average of the supplied "
                        f"anisotropic dielectric. This could lead to significant errors "
                        f"for very anisotropic systems and/or relatively small supercells!"
                    )
            except Exception as freysoldt_exc:
                warnings.warn(
                    f"Got this error message when attempting to parse defect & bulk "
                    f"`LOCPOT` files to compute the Freysoldt (FNV) charge correction:"
                    f"\n{freysoldt_exc}\n-> Charge corrections will not be applied for this "
                    f"defect."
                )
                if not isotropic_dielectric:
                    # reset dielectric to original anisotropic value if FNV failed as well:
                    sdp.defect_entry.parameters["dielectric"] = dielectric
                skip_corrections = True

        else:
            if int(sdp.defect_entry.charge) != 0:
                warnings.warn(
                    f"`LOCPOT` or `OUTCAR` files are not present in both the defect (at "
                    f"{defect_path}) and bulk (at {bulk_path}) folders. These are needed to "
                    f"perform the finite-size charge corrections. Charge corrections will not be "
                    f"applied for this defect."
                )
                skip_corrections = True

        if not skip_corrections:
            # Check compatibility of defect corrections with loaded metadata, and apply
            sdp.run_compatibility()

    return sdp.defect_entry


def dpd_from_defect_dict(parsed_defect_dict: dict) -> DefectPhaseDiagram:
    """Generates a DefectPhaseDiagram object from a dictionary of parsed defect calculations (
    format: {"defect_name": defect_entry}), likely created using SingleDefectParser from
    doped.pycdt.utils.parse_calculations), which can then be used to analyse and plot the defect
    thermodynamics (formation energies, transition levels, concentrations etc)

    Args:
        parsed_defect_dict (dict):
            Dictionary of parsed defect calculations (format: {"defect_name": defect_entry}),
            likely created using SingleDefectParser from doped.pycdt.utils.parse_calculations).
            Must have 'vbm' and 'gap' in defect_entry.parameters for each defect (from
            SingleDefectParser.get_bulk_gap_data())

    Returns:
        pymatgen DefectPhaseDiagram object (DefectPhaseDiagram)
    """
    # TODO: Can we make the dpd generation more efficient? What's the bottleneck in it's
    #  initialisation? `pymatgen` site-matching that can be avoided?
    # TODO: Write our own DefectPhaseDiagram class, to (1) refactor the site-matching to just
    #  use the already-parsed site positions, and then merge interstitials according to this
    #  algorithm:
    # 1. For each interstitial defect type, count the number of parsed calculations per charge
    # state, and take the charge state with the most calculations present as our starting point (
    # if multiple charge states have the same number of calculations, take the closest to neutral).
    # 2. For each interstitial in a different charge state, determine which of the starting
    # points has their (already-parsed) Voronoi site closest to its (already-parsed) Voronoi
    # site, making sure to account for symmetry equivalency (using just Voronoi sites + bulk
    # structure will be easiest), and merge with this.
    # Also add option to just amalgamate and show only the lowest energy states.
    #  (2) optionally retain/remove unstable (in the gap) charge states (rather than current
    #  default range of (VBM - 1eV, CBM + 1eV))...
    vbm_vals = []
    bandgap_vals = []
    for defect in parsed_defect_dict.values():
        vbm_vals.append(defect.parameters["vbm"])
        bandgap_vals.append(defect.parameters["gap"])
    if len(set(vbm_vals)) > 1:  # Check if all defects give same vbm
        raise ValueError(
            f"VBM values don't match for defects in given defect dictionary, "
            f"the VBM values in the dictionary are: {vbm_vals}. "
            f"Are you sure the correct/same bulk files were used with "
            f"SingleDefectParser and/or get_bulk_gap_data()?"
        )
    if len(set(bandgap_vals)) > 1:  # Check if all defects give same bandgap
        raise ValueError(
            f"Bandgap values don't match for defects in given defect dictionary, "
            f"the bandgap values in the dictionary are: {bandgap_vals}. "
            f"Are you sure the correct/same bulk files were used with "
            f"SingleDefectParser and/or get_bulk_gap_data()?"
        )
    vbm = vbm_vals[0]
    bandgap = bandgap_vals[0]
    dpd = DefectPhaseDiagram(
        list(parsed_defect_dict.values()), vbm, bandgap, filter_compatible=False
    )

    return dpd


def dpd_transition_levels(defect_phase_diagram: DefectPhaseDiagram):
    """Iteratively prints the charge transition levels for the input DefectPhaseDiagram object
    (via the from a defect_phase_diagram.transition_level_map attribute)

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
            DefectPhaseDiagram object (likely created from
            analysis.dpd_from_defect_dict)

    Returns:
        None
    """
    for def_type, tl_info in defect_phase_diagram.transition_level_map.items():
        bold_print(f"\nDefect: {def_type.split('@')[0]}")
        for tl_efermi, chargeset in tl_info.items():
            print(
                f"Transition Level ({max(chargeset):{'+' if max(chargeset) else ''}}/"
                f"{min(chargeset):{'+' if min(chargeset) else ''}}) at {tl_efermi:.3f}"
                f" eV above the VBM"
            )


def formation_energy_table(
    defect_phase_diagram: DefectPhaseDiagram,
    chempot_limits: dict = None,
    pd_facets: list = None,
    fermi_level: float = 0,
    hide_cols: list = None,
    show_key: bool = True,
):
    """
    Prints defect formation energy tables for either a single chemical potential limit (i.e. phase
    diagram facet) or each facet in the phase diagram (chempot_limits dict), depending on the
    chempot_limits input supplied. This can either be a dictionary of chosen absolute/DFT chemical
    potentials: {Elt: Energy} (giving a single formation energy table) or a dictionary including
    the key-value pair: {"facets": [{'facet': [chempot_dict]}]}, following the format generated
    by chempot_limits = cpa.read_phase_diagram_and_chempots() (see example notebooks). In the
    latter case, a subset of facet(s) / chemical potential limit(s) can be chosen with the
    pd_facets argument, or if not specified, will print formation energy tables for each facet in
    the phase diagram.
    Returns the results a pandas DataFrame or list of DataFrames.

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
             DefectPhaseDiagram object (likely created from
             analysis.dpd_from_defect_dict)
        chempot_limits (dict):
            This can either be a dictionary of chosen absolute/DFT chemical potentials: {Elt:
            Energy} (giving a single formation energy table) or a dictionary including the
            key-value pair: {"facets": [{'facet': [chempot_dict]}]}, following the format generated
            by chempot_limits = cpa.read_phase_diagram_and_chempots() (see example notebooks). If
            not specified, chemical potentials are not included in the formation energy calculation
            (all set to zero energy).
        pd_facets (list):
            A list facet(s) / chemical potential limit(s) for which to print the defect formation
            energy tables. If not specified, will print formation energy tables for each facet in
            the phase diagram. (default: None)
        fermi_level (float):
            Fermi level to use for computing the defect formation energies. (default: 0 (i.e.
            at the VBM))
        hide_cols: (list):
            List of columns to hide from the output. (default: None)
        show_key (bool):
            Whether or not to print the table key at the bottom of the output. (default: True)

    Returns:
        pandas DataFrame or list of DataFrames
    """
    if chempot_limits is None:
        chempot_limits = {}

    if "facets" in chempot_limits:
        list_of_dfs = []
        if not pd_facets:
            pd_facets = chempot_limits[
                "facets"
            ].keys()  # Phase diagram facets to use for chemical
            # potentials, to tabulate formation energies
        for facet in pd_facets:
            bold_print("Facet: " + unicodeify(facet))
            df = single_formation_energy_table(
                defect_phase_diagram,
                chempots=chempot_limits["facets"][facet],
                fermi_level=fermi_level,
                hide_cols=hide_cols,
                show_key=show_key,
            )
            list_of_dfs.append(df)
            print("\n")

        return list_of_dfs

    # else return {Elt: Energy} dict for chempot_limits, or if unspecified, all zero energy
    df = single_formation_energy_table(
        defect_phase_diagram,
        chempots=chempot_limits,
        fermi_level=fermi_level,
        hide_cols=hide_cols,
        show_key=show_key,
    )
    return df


def single_formation_energy_table(
    defect_phase_diagram: DefectPhaseDiagram,
    chempots: dict = None,
    fermi_level: float = 0,
    hide_cols: list = None,
    show_key: bool = True,
):
    """
    Prints a defect formation energy table for a single chemical potential limit (i.e. phase diagram
    facet), and returns the results as a pandas DataFrame.

    Args:
        defect_phase_diagram (DefectPhaseDiagram):
             DefectPhaseDiagram object (likely created from
             analysis.dpd_from_defect_dict)
        chempots (dict):
            Dictionary of chosen absolute/DFT chemical potentials: {Elt: Energy}. If not
            specified, chemical potentials are not included in the formation energy calculation
            (all set to zero energy).
        fermi_level (float):
            Fermi level to use for computing the defect formation energies. (default: 0 (i.e.
            at the VBM))
        hide_cols: (list):
            List of columns to hide from the output. (default: None)
        show_key (bool):
            Whether or not to print the table key at the bottom of the output. (default: True)

    Returns:
        pandas DataFrame sorted by formation energy
    """
    header = ["Defect", "Charge", "Defect Path"]
    table = []
    if hide_cols is None:
        hide_cols = []

    for defect_entry in defect_phase_diagram.entries:
        row = [
            defect_entry.name,
            defect_entry.charge,
            defect_entry.parameters["defect_path"],
        ]
        if "Uncorrected Energy" not in hide_cols:
            header += ["Uncorrected Energy"]
            row += [f"{defect_entry.uncorrected_energy:.2f} eV"]
        if "Corrected Energy" not in hide_cols:
            header += ["Corrected Energy"]
            row += [
                f"{defect_entry.energy:.2f} eV"
            ]  # With 0 chemical potentials, at the calculation
            # fermi level
        header += ["Formation Energy"]
        formation_energy = defect_entry.formation_energy(
            chemical_potentials=chempots, fermi_level=fermi_level
        )
        row += [f"{formation_energy:.2f} eV"]

        table.append(row)
    table = sorted(table, key=itemgetter(0, 1))
    print(
        tabulate(
            table,
            headers=header,
            tablefmt="fancy_grid",
            stralign="left",
            numalign="left",
        ),
        "\n",
    )

    if show_key:
        bold_print("Table Key:")
        print(
            """'Defect' -> Defect Type and Multiplicity
'Charge' -> Defect Charge State
'Uncorrected Energy' -> Defect Energy from calculation, without corrections
'Corrected Energy' -> Defect Energy from calculation (E_defect - E_host + corrections)
(chemical potentials set to 0 and the fermi level at average electrostatic potential in the
supercell)
'Formation Energy' -> Final Defect Formation Energy, with the specified chemical potentials (
chempot_limits)(default: all 0) and the chosen fermi_level (default: 0)(i.e. at the VBM)
        """
        )

    sorted_df = pd.DataFrame(
        table,
        columns=[
            "Defect",
            "Charge",
            "Defect Path",
            "Uncorrected Energy",
            "Corrected Energy",
            "Formation Energy",
        ],
    )
    sorted_df = sorted_df.sort_values("Formation Energy")
    return sorted_df


def lany_zunger_corrected_defect_dict_from_freysoldt(defect_dict: dict):
    """Convert input parsed defect dictionary (presumably created using SingleDefectParser
    from doped.pycdt.utils.parse_calculations) with Freysoldt charge corrections to
    the same parsed defect dictionary but with the Lany-Zunger charge correction (same potential
    alignment plus 0.65 * Makov-Payne image charge correction (same image charge correction as
    Freysoldt scheme)).
    Args:
        parsed_defect_dict (dict):
            Dictionary of parsed defect calculations (presumably created using SingleDefectParser
            from doped.pycdt.utils.parse_calculations) (see example notebook)
            Must have 'freysoldt_meta' in defect.parameters for each charged defect (from
            SingleDefectParser.freysoldt_loader())
    Returns:
        Parsed defect dictionary with Lany-Zunger charge corrections.
    """
    from doped.corrections import (
        get_murphy_image_charge_correction,
    )  # avoid circular import

    random_defect_entry = list(defect_dict.values())[
        0
    ]  # Just need any DefectEntry from
    # defect_dict to get the lattice and dielectric matrix
    lattice = random_defect_entry.bulk_structure.lattice.matrix
    dielectric = random_defect_entry.parameters["dielectric"]
    lz_image_charge_corrections = get_murphy_image_charge_correction(
        lattice, dielectric
    )
    lz_corrected_defect_dict = copy.deepcopy(defect_dict)
    for defect_name, defect_entry in lz_corrected_defect_dict.items():
        if defect_entry.charge != 0:
            potalign = defect_entry.parameters["freysoldt_meta"][
                "freysoldt_potential_alignment_correction"
            ]
            mp_pc_corr = lz_image_charge_corrections[
                abs(defect_entry.charge)
            ]  # Makov-Payne PC correction
            defect_entry.parameters.update(
                {
                    "Lany-Zunger_Corrections": {
                        "(Freysoldt)_Potential_Alignment_Correction": potalign,
                        "Makov-Payne_Image_Charge_Correction": mp_pc_corr,
                        "Lany-Zunger_Scaled_Image_Charge_Correction": 0.65 * mp_pc_corr,
                        "Total_Lany-Zunger_Correction": potalign + 0.65 * mp_pc_corr,
                    }
                }
            )
            defect_entry.corrections["charge_correction"] = defect_entry.parameters[
                "Lany-Zunger_Corrections"
            ]["Total_Lany-Zunger_Correction"]

        lz_corrected_defect_dict.update({defect_name: defect_entry})
    return lz_corrected_defect_dict


def lany_zunger_corrected_defect_dict_from_kumagai(defect_dict: dict):
    """Convert input parsed defect dictionary (presumably created using SingleDefectParser
    from doped.pycdt.utils.parse_calculations) with Kumagai charge corrections to
    the same parsed defect dictionary but with the 'Lany-Zunger' charge correction (same potential
    alignment plus 0.65 * image charge correction.
    Args:
        parsed_defect_dict (dict):
            Dictionary of parsed defect calculations (presumably created using SingleDefectParser
            from doped.pycdt.utils.parse_calculations) (see example notebook)
            Must have 'kumagai_meta' in defect.parameters for each charged defect (from
            SingleDefectParser.kumagai_loader())
    Returns:
        Parsed defect dictionary with Lany-Zunger charge corrections.
    """
    from doped.corrections import (
        get_murphy_image_charge_correction,
    )  # avoid circular import

    random_defect_entry = list(defect_dict.values())[
        0
    ]  # Just need any DefectEntry from
    # defect_dict to get the lattice and dielectric matrix
    lattice = random_defect_entry.bulk_structure.lattice.matrix
    dielectric = random_defect_entry.parameters["dielectric"]
    lz_image_charge_corrections = get_murphy_image_charge_correction(
        lattice, dielectric
    )
    lz_corrected_defect_dict = copy.deepcopy(defect_dict)
    for defect_name, defect_entry in lz_corrected_defect_dict.items():
        if defect_entry.charge != 0:
            potalign = defect_entry.parameters["kumagai_meta"][
                "kumagai_potential_alignment_correction"
            ]
            makove_payne_pc_correction = lz_image_charge_corrections[
                abs(defect_entry.charge)
            ]
            defect_entry.parameters.update(
                {
                    "Lany-Zunger_Corrections": {
                        "(Kumagai)_Potential_Alignment_Correction": potalign,
                        "Makov-Payne_Image_Charge_Correction": makove_payne_pc_correction,
                        "Lany-Zunger_Scaled_Image_Charge_Correction": 0.65
                        * makove_payne_pc_correction,
                        "Total_Lany-Zunger_Correction": potalign
                        + 0.65 * makove_payne_pc_correction,
                    }
                }
            )
            defect_entry.corrections["charge_correction"] = defect_entry.parameters[
                "Lany-Zunger_Corrections"
            ]["Total_Lany-Zunger_Correction"]

        lz_corrected_defect_dict.update({defect_name: defect_entry})
    return lz_corrected_defect_dict



# coding: utf-8

from __future__ import division

__status__ = "Development"

import copy
import inspect
import os
import unittest
from shutil import copyfile

from monty.serialization import loadfn
from monty.tempfile import ScratchDir
from pymatgen.analysis.defects.thermodynamics import DefectPhaseDiagram
from pymatgen.analysis.phase_diagram import PhaseDiagram
from pymatgen.core import Composition, Element
from pymatgen.entries.computed_entries import ComputedEntry
from pymatgen.ext.matproj import MPRester
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.core._chemical_potentials import (
    ChemPotAnalyzer,
    MPChemPotAnalyzer,
    UserChemPotAnalyzer,
    UserChemPotInputGenerator,
    get_mp_chempots_from_dpd,
)

TEST_DIR = os.path.abspath(os.path.join(__file__, "..", "..", "..", "test_files"))


class UserChemPotAnalyzerTest(PymatgenTest):
    def setUp(self):
        with MPRester(api_key="c2LiJRMiBeaN5iXsH") as mp:
            self.bulk_ce = mp.get_entry_by_material_id("mp-2534")
        self.UCPA = UserChemPotAnalyzer(
            bulk_ce=self.bulk_ce, mapi_key="c2LiJRMiBeaN5iXsH"
        )
        # SK MP Imperial email A/C API key
        self.UCPA_sub = UserChemPotAnalyzer(
            bulk_ce=self.bulk_ce, sub_species=["In"], mapi_key="c2LiJRMiBeaN5iXsH"
        )

    def test_read_phase_diagram_and_chempots(self):
        # set up a local phase diagram object...
        # test non mp case,
        with ScratchDir("."):
            # os.mkdir('PhaseDiagram')
            os.makedirs(os.path.join("PhaseDiagram", "Ga"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_Ga"),
                os.path.join("PhaseDiagram", "Ga", "vasprun.xml"),
            )
            os.mkdir(os.path.join("PhaseDiagram", "As"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_As"),
                os.path.join("PhaseDiagram", "As", "vasprun.xml"),
            )
            os.mkdir(os.path.join("PhaseDiagram", "GaAs"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_GaAs"),
                os.path.join("PhaseDiagram", "GaAs", "vasprun.xml"),
            )
            cp = self.UCPA.read_phase_diagram_and_chempots(
                full_sub_approach=False, include_mp_entries=False
            )
            self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(cp["facets"].keys()))
            self.assertEqual(
                [-5.36, -4.29],
                [
                    round(cp["facets"]["As-GaAs"][Element(elt)], 2)
                    for elt in ["As", "Ga"]
                ],
            )
            self.assertEqual(
                [-6.04, -3.61],
                [
                    round(cp["facets"]["Ga-GaAs"][Element(elt)], 2)
                    for elt in ["As", "Ga"]
                ],
            )

        # followed by an case where MP needs to supplement...
        with ScratchDir("."):
            os.mkdir("PhaseDiagram")
            # NO Ga entry included this time
            os.mkdir(os.path.join("PhaseDiagram", "As"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_As"),
                os.path.join("PhaseDiagram", "As", "vasprun.xml"),
            )
            os.mkdir(os.path.join("PhaseDiagram", "GaAs"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_GaAs"),
                os.path.join("PhaseDiagram", "GaAs", "vasprun.xml"),
            )
            cp = self.UCPA.read_phase_diagram_and_chempots(
                full_sub_approach=False, include_mp_entries=True
            )
            self.assertEqual(set(["As-GaAs", "Ga-GaAs"]), set(cp["facets"].keys()))
            self.assertEqual(
                [-5.36, -4.29],
                [
                    round(cp["facets"]["As-GaAs"][Element(elt)], 2)
                    for elt in ["As", "Ga"]
                ],
            )
            self.assertEqual(
                [-6.62, -3.03],
                [
                    round(cp["facets"]["Ga-GaAs"][Element(elt)], 2)
                    for elt in ["As", "Ga"]
                ],
            )

        # quick and dirty test for finding extrinsic defects...
        with ScratchDir("."):
            os.mkdir("PhaseDiagram")
            # NO Ga entry or In entry this time
            os.mkdir(os.path.join("PhaseDiagram", "As"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_As"),
                os.path.join("PhaseDiagram", "As", "vasprun.xml"),
            )
            os.mkdir(os.path.join("PhaseDiagram", "GaAs"))
            copyfile(
                os.path.join(TEST_DIR, "vasprun.xml_GaAs"),
                os.path.join("PhaseDiagram", "GaAs", "vasprun.xml"),
            )
            cp = self.UCPA_sub.read_phase_diagram_and_chempots(
                full_sub_approach=False, include_mp_entries=True)
            self.assertEqual({'As-GaAs-In', 'Ga-GaAs-In'}, set(cp["facets"].keys()))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

from __future__ import print_function

__status__ = "Development"

import os
import warnings

import numpy as np

from monty.os.path import zpath
from monty.serialization import loadfn
from pymatgen.io.vasp.inputs import Kpoints, Potcar, PotcarSingle
from pymatgen.io.vasp.sets import MPRelaxSet, MPStaticSet, DictSet


def _check_psp_dir():  # Provided by Katarina Brlec, from github.com/SMTG-UCL/surfaxe
    """
    Helper function to check if potcars are set up correctly for use with
    pymatgen, to be compatible across pymatgen versions (breaking changes in v2022)
    """
    potcar = False
    try:
        import pymatgen.settings

        pmg_settings = pymatgen.settings.SETTINGS
        if "PMG_VASP_PSP_DIR" in pmg_settings:
            potcar = True
    except ModuleNotFoundError:
        try:
            import pymatgen

            pmg_settings = pymatgen.SETTINGS
            if "PMG_VASP_PSP_DIR" in pmg_settings:
                potcar = True
        except AttributeError:
            from pymatgen.core import SETTINGS

            pmg_settings = SETTINGS
            if "PMG_VASP_PSP_DIR" in pmg_settings:
                potcar = True
    return potcar


def _import_psp():
    """import pmg settings for PotcarSingleMod"""
    pmg_settings = None
    try:
        import pymatgen.settings

        pmg_settings = pymatgen.settings.SETTINGS
    except ModuleNotFoundError:
        try:
            import pymatgen

            pmg_settings = pymatgen.SETTINGS
        except AttributeError:
            from pymatgen.core import SETTINGS

            pmg_settings = SETTINGS

    if pmg_settings is None:
        raise ValueError("pymatgen settings not found?")
    else:
        return pmg_settings


class PotcarSingleMod(PotcarSingle):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

    @staticmethod
    def from_symbol_and_functional(symbol, functional=None):
        settings = _import_psp()
        if functional is None:
            functional = settings.get("PMG_DEFAULT_FUNCTIONAL", "PBE_54")
        funcdir = PotcarSingle.functional_dir[functional]

        if not os.path.isdir(os.path.join(settings.get("PMG_VASP_PSP_DIR"), funcdir)):
            functional_dir = {
                "LDA_US": "pot",
                "PW91_US": "pot_GGA",
                "LDA": "potpaw",
                "PW91": "potpaw_GGA",
                "LDA_52": "potpaw_LDA.52",
                "LDA_54": "potpaw_LDA.54",
                "PBE": "potpaw_PBE",
                "PBE_52": "potpaw_PBE.52",
                "PBE_54": "potpaw_PBE.54",
            }
            funcdir = functional_dir[functional]

        d = settings.get("PMG_VASP_PSP_DIR")
        if d is None:
            raise ValueError(
                "No POTCAR directory found. Please set "
                "the VASP_PSP_DIR environment variable"
            )

        paths_to_try = [
            os.path.join(d, funcdir, "POTCAR.{}".format(symbol)),
            os.path.join(d, funcdir, symbol, "POTCAR.Z"),
            os.path.join(d, funcdir, symbol, "POTCAR"),
        ]
        for p in paths_to_try:
            p = os.path.expanduser(p)
            p = zpath(p)
            if os.path.exists(p):
                return PotcarSingleMod.from_file(p)
        raise IOError(
            "You do not have the right POTCAR with functional "
            + "{} and label {} in your VASP_PSP_DIR".format(functional, symbol)
        )


class PotcarMod(Potcar):
    def __init__(self, **kwargs):
        super(self.__class__, self).__init__(**kwargs)

    def set_symbols(self, symbols, functional=None, sym_potcar_map=None):
        """
        Initialize the POTCAR from a set of symbols. Currently, the POTCARs can
        be fetched from a location specified in .pmgrc.yaml. Use pmg config
        to add this setting.

        Args:
            symbols ([str]): A list of element symbols
            functional (str): The functional to use. If None, the setting
                PMG_DEFAULT_FUNCTIONAL in .pmgrc.yaml is used, or if this is
                not set, it will default to PBE_54.
            sym_potcar_map (dict): A map of symbol:raw POTCAR string. If
                sym_potcar_map is specified, POTCARs will be generated from
                the given map data rather than the config file location.
        """
        del self[:]
        if sym_potcar_map:
            for el in symbols:
                self.append(PotcarSingleMod(sym_potcar_map[el]))
        else:
            for el in symbols:
                p = PotcarSingleMod.from_symbol_and_functional(el, functional)
                self.append(p)


class DefectRelaxSet(DictSet):
    """
    pymatgen DictSet for VASP Defect Relaxation Calculations
    Additional Args:
        charge: Charge of the defect structure
    """

    def __init__(self, structure, **kwargs):
        charge = kwargs.pop("charge", 0)
        config_dict = kwargs.pop("config_dict", {})
        if config_dict:
            super(self.__class__, self).__init__(
                structure, config_dict=config_dict, **kwargs
            )
        else:
            MODULE_DIR = os.path.dirname(os.path.abspath(__file__))
            default_potcar_dict = loadfn(
                os.path.join(MODULE_DIR, "../../PotcarSet.yaml")
            )
            relax_set = loadfn(os.path.join(MODULE_DIR, "../../HSE06_RelaxSet.yaml"))
            defect_set = loadfn(os.path.join(MODULE_DIR, "../../DefectSet.yaml"))
            relax_set["INCAR"].update(defect_set["INCAR"])
            relax_set.update(default_potcar_dict)

            from doped.vasp_input import scaled_ediff

            relax_set["INCAR"]["EDIFF"] = scaled_ediff(len(structure))

            super(self.__class__, self).__init__(
                structure, config_dict=relax_set, **kwargs
            )

        self.charge = charge

    @property
    def incar(self):
        inc = super(self.__class__, self).incar
        try:
            inc["NELECT"] = self.nelect - self.charge
            if inc["NELECT"] % 2 != 0:  # odd number of electrons
                inc["NUPDOWN"] = 1
            else:
                inc[
                    "NUPDOWN"
                ] = 0  # But could be 2 for triplet states (e.g. bipolarons)

        except Exception as e:
            warnings.warn(
                f"NELECT and NUPDOWN flags are not set due to non-availability of POTCARs; "
                f"got error {e}"
            )

        return inc

    @property
    def potcar(self):
        """
        Potcar object.
        """
        return PotcarMod(symbols=self.potcar_symbols, functional=self.potcar_functional)

    @property
    def all_input(self):
        """
        Returns all input files as a dict of {filename: vasp object}

        Returns:
            dict of {filename: object}, e.g., {'INCAR': Incar object, ...}
        """
        try:
            return super(DefectRelaxSet, self).all_input
        except:  # Expecting the error to be POTCAR related, its ignored
            kpoints = self.kpoints
            incar = self.incar
            if np.product(kpoints.kpts) < 4 and incar.get("ISMEAR", 0) == -5:
                incar["ISMEAR"] = 0

            return {"INCAR": incar, "KPOINTS": kpoints, "POSCAR": self.poscar}

    @property
    def structure(self):
        """
        :return: Structure
        """
        return self._structure



# coding: utf-8

from __future__ import division

__status__ = "Development"

import os
import tarfile
import unittest
from shutil import copyfile
import numpy as np

from monty.tempfile import ScratchDir
from pymatgen.analysis.defects.core import DefectEntry, Substitution, Vacancy
from pymatgen.core import Element, PeriodicSite
from pymatgen.io.vasp import Locpot, Vasprun
from pymatgen.util.testing import PymatgenTest

from doped.pycdt.utils.parse_calculations import PostProcess, SingleDefectParser

file_loc = os.path.abspath(os.path.join(__file__, "..", "..", "..", "test_files"))


class SingleDefectParserTest(PymatgenTest):
    def test_all_methods(self):
        # create scratch directory with files....
        # having to do it all at once to minimize amount of time copying over to Scratch Directory
        with ScratchDir("."):
            # setup with fake Locpot object copied over
            copyfile(
                os.path.join(file_loc, "test_path_files.tar.gz"),
                "./test_path_files.tar.gz",
            )
            tar = tarfile.open("test_path_files.tar.gz")
            tar.extractall()
            tar.close()
            blocpot = Locpot.from_file(os.path.join(file_loc, "bLOCPOT.gz"))
            blocpot.write_file("test_path_files/bulk/LOCPOT")
            dlocpot = Locpot.from_file(os.path.join(file_loc, "dLOCPOT.gz"))
            dlocpot.write_file("test_path_files/sub_1_Sb_on_Ga/charge_2/LOCPOT")

            # test_from_path
            sdp = SingleDefectParser.from_paths(
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
                "test_path_files/bulk/",
                18.12,
                charge=2,
            )
            self.assertIsInstance(sdp, SingleDefectParser)
            self.assertIsInstance(sdp.defect_entry.defect, Substitution)
            self.assertEqual(
                sdp.defect_entry.parameters["bulk_path"], "test_path_files/bulk/"
            )
            self.assertEqual(
                sdp.defect_entry.parameters["defect_path"],
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
            )

            # test_freysoldt_loader
            for param_key in [
                "axis_grid",
                "bulk_planar_averages",
                "defect_planar_averages",
                "defect_frac_sc_coords",
            ]:
                self.assertFalse(param_key in sdp.defect_entry.parameters.keys())
            sdp.freysoldt_loader()
            for param_key in [
                "axis_grid",
                "bulk_planar_averages",
                "defect_planar_averages",
                "initial_defect_structure",
                "defect_frac_sc_coords",
            ]:
                self.assertTrue(param_key in sdp.defect_entry.parameters.keys())

            # test_kumagai_loader
            for param_key in [
                "bulk_atomic_site_averages",
                "defect_atomic_site_averages",
                "site_matching_indices",
                "sampling_radius",
            ]:
                self.assertFalse(param_key in sdp.defect_entry.parameters.keys())
            sdp.kumagai_loader()
            for param_key in [
                "bulk_atomic_site_averages",
                "defect_atomic_site_averages",
                "site_matching_indices",
                "defect_frac_sc_coords",
            ]:
                self.assertTrue(param_key in sdp.defect_entry.parameters.keys())

            # test_get_stdrd_metadata
            sdp.get_stdrd_metadata()
            for param_key in [
                "eigenvalues",
                "kpoint_weights",
                "bulk_energy",
                "final_defect_structure",
                "defect_energy",
                "run_metadata",
            ]:
                self.assertTrue(param_key in sdp.defect_entry.parameters.keys())

            # test_get_bulk_gap_data
            sdp.get_bulk_gap_data()
            self.assertAlmostEqual(sdp.defect_entry.parameters["gap"], 1.2167)

            # test_run_compatibility
            self.assertFalse("is_compatible" in sdp.defect_entry.parameters)
            sdp.run_compatibility()
            self.assertTrue("is_compatible" in sdp.defect_entry.parameters)


class PostProcessTest(PymatgenTest):
    def test_get_chempot_limits(self):
        with ScratchDir("."):
            # make a fake file structure to parse vaspruns
            os.mkdir("bulk")
            copyfile(os.path.join(file_loc, "vasprun.xml_GaAs"), "bulk/vasprun.xml")
            pp = PostProcess(".", mapi_key="c2LiJRMiBeaN5iXsH")  # SK MP Imperial email A/C API key
            gaas_cp = pp.get_chempot_limits()
            self.assertEqual({"As-GaAs", "Ga-GaAs"}, set(gaas_cp.keys()))
            np.testing.assert_almost_equal(
                [-4.65807055, -4.9884807425],
                [gaas_cp["As-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
                decimal=3)
            np.testing.assert_almost_equal(
                [-6.618, -3.028],
                [gaas_cp["Ga-GaAs"][Element(elt)] for elt in ["As", "Ga"]],
                decimal=3)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8

from __future__ import division

import unittest
import os

from pymatgen.core.structure import Structure
from monty.serialization import loadfn

from doped.pycdt.utils.vasp import PotcarMod, PotcarSingleMod, DefectRelaxSet

__status__ = "Development"

file_loc = os.path.abspath(os.path.join(__file__, "..", "..", "..", "test_files"))


class DefectRelaxTest(unittest.TestCase):
    def setUp(self):
        self.structure = Structure.from_file(os.path.join(file_loc, "POSCAR_Cr2O3"))
        self.user_settings = loadfn(os.path.join(file_loc, "test_vasp_settings.yaml"))
        self.path = "Cr2O3"
        self.neutral_def_incar_min = {
              "ICORELEVEL": "0  # Needed if using the Kumagai-Oba (eFNV) anisotropic charge "
                            "correction scheme".lower(),
              "ISIF": 2,  # Fixed supercell for defects
              "ISPIN": 2,  # Spin polarisation likely for defects
              "ISYM": "0  # Symmetry breaking extremely likely for defects".lower(),
              "LVHAR": True,
              "ISMEAR": 0,
        }
        self.def_keys = ["EDIFF", "EDIFFG", "IBRION"]

    def test_neutral_defect_incar(self):
        drs = DefectRelaxSet(self.structure)
        self.assertTrue(self.neutral_def_incar_min.items() <= drs.incar.items())
        self.assertTrue(set(self.def_keys).issubset(drs.incar))

    def test_charged_defect_incar(self):
        drs = DefectRelaxSet(self.structure, charge=1)
        self.assertTrue(self.neutral_def_incar_min.items() <= drs.incar.items())
        self.assertTrue(set(self.def_keys).issubset(drs.incar))

    def test_user_settings_defect_incar(self):
        user_incar_settings = {"EDIFF": 1e-8, "EDIFFG": 0.1, "ENCUT": 720}
        drs = DefectRelaxSet(
            self.structure, charge=1, user_incar_settings=user_incar_settings
        )
        self.assertTrue(self.neutral_def_incar_min.items() <= drs.incar.items())
        self.assertTrue(set(self.def_keys).issubset(drs.incar))
        self.assertEqual(drs.incar["ENCUT"], 720)
        self.assertEqual(drs.incar["EDIFF"], 1e-8)
        self.assertEqual(drs.incar["EDIFFG"], 0.1)


if __name__ == "__main__":
    unittest.main()



# coding: utf-8
"""
Parses the computed data from VASP defect calculations.
"""

import glob
import os
import warnings

import numpy as np
from monty.json import MontyDecoder
from monty.serialization import dumpfn, loadfn
from pymatgen.analysis.defects.core import (
    DefectEntry,
    Interstitial,
    Substitution,
    Vacancy,
)
from pymatgen.analysis.defects.defect_compatibility import DefectCompatibility
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.core import PeriodicSite, Structure
from pymatgen.entries.computed_entries import ComputedStructureEntry
from pymatgen.ext.matproj import MPRester
from pymatgen.io.vasp.inputs import Potcar, UnknownPotcarWarning
from pymatgen.io.vasp.outputs import Locpot, Outcar, Poscar, Vasprun
from pymatgen.util.coord import pbc_diff

from doped.pycdt.core import _chemical_potentials
from doped import _ignore_pmg_warnings

_ignore_pmg_warnings()


def _custom_formatwarning(msg, *args, **kwargs):
    """Reformat warnings to just print the warning message"""
    return f"{msg}\n"


warnings.formatwarning = _custom_formatwarning


def get_vasprun(vasprun_path, **kwargs):
    """Read the vasprun.xml(.gz) file as a pymatgen Vasprun object"""
    vasprun_path = str(vasprun_path)  # convert to string if Path object
    warnings.filterwarnings(
        "ignore", category=UnknownPotcarWarning
    )  # Ignore POTCAR warnings when loading vasprun.xml
    # pymatgen assumes the default PBE with no way of changing this within get_vasprun())
    warnings.filterwarnings(
        "ignore", message="No POTCAR file with matching TITEL fields"
    )
    if os.path.exists(vasprun_path) and os.path.isfile(vasprun_path):
        vasprun = Vasprun(vasprun_path, **kwargs)
    else:
        raise FileNotFoundError(
            f"vasprun.xml file not found at {vasprun_path}. Needed for parsing calculation output."
        )
    return vasprun


def get_locpot(locpot_path):
    """Read the LOCPOT(.gz) file as a pymatgen Locpot object"""
    locpot_path = str(locpot_path)  # convert to string if Path object
    if os.path.exists(locpot_path) and os.path.isfile(locpot_path):
        locpot = Locpot.from_file(locpot_path)
    else:
        raise FileNotFoundError(
            f"LOCPOT file not found at {locpot_path}. Needed for calculating the Freysoldt (FNV) "
            f"image charge correction."
        )
    return locpot


def get_outcar(outcar_path):
    """Read the OUTCAR(.gz) file as a pymatgen Outcar object"""
    outcar_path = str(outcar_path)  # convert to string if Path object
    if os.path.exists(outcar_path) and os.path.isfile(outcar_path):
        outcar = Outcar(outcar_path)
    else:
        raise FileNotFoundError(
            f"OUTCAR file not found at {outcar_path}. Needed for calculating the Kumagai (eFNV) "
            f"image charge correction."
        )
    return outcar


def _get_output_files_and_check_if_multiple(output_file="vasprun.xml", path="."):
    """
    Search for all files with filenames matching `output_file`, case-insensitive.

    Returns (output file path, Multiple?) where Multiple is True if multiple matching files are
    found.
    """
    files = os.listdir(path)
    output_files = [
        filename for filename in files if output_file.lower() in filename.lower()
    ]
    # sort by, direct match to output_file, direct match to output_file with .gz extension,
    # then alphabetically:
    output_files = sorted(
        output_files,
        key=lambda x: (
            x == output_file,
            x == output_file + ".gz",
            x,
        ),
        reverse=True,
    )
    if output_files:
        output_path = os.path.join(path, output_files[0])
        if len(output_files) > 1:
            return output_path, True
        else:
            return output_path, False
    else:
        return path, False  # so when `get_X()` is called, it will raise an
        # informative FileNotFoundError


def get_defect_type_and_composition_diff(bulk, defect):
    """Get the difference in composition between a bulk structure and a defect structure.
    Contributed by Dr. Alex Ganose (@ Imperial Chemistry) and refactored for extrinsic species
    """
    bulk_comp = bulk.composition.get_el_amt_dict()
    defect_comp = defect.composition.get_el_amt_dict()

    composition_diff = {
        element: int(defect_amount - bulk_comp.get(element, 0))
        for element, defect_amount in defect_comp.items()
        if int(defect_amount - bulk_comp.get(element, 0)) != 0
    }

    if len(composition_diff) == 1 and list(composition_diff.values())[0] == 1:
        defect_type = "interstitial"
    elif len(composition_diff) == 1 and list(composition_diff.values())[0] == -1:
        defect_type = "vacancy"
    elif len(composition_diff) == 2:
        defect_type = "substitution"
    else:
        raise RuntimeError(
            "Could not determine defect type from composition difference of bulk "
            "and defect structures."
        )

    return defect_type, composition_diff


def get_defect_site_idxs_and_unrelaxed_structure(
    bulk, defect, defect_type, composition_diff, unique_tolerance=1
):
    """Get the defect site and unrelaxed structure.
    Contributed by Dr. Alex Ganose (@ Imperial Chemistry) and refactored for extrinsic species
    """
    if defect_type == "substitution":
        old_species = [el for el, amt in composition_diff.items() if amt == -1][0]
        new_species = [el for el, amt in composition_diff.items() if amt == 1][0]

        bulk_new_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == new_species]
        )
        defect_new_species_coords = np.array(
            [site.frac_coords for site in defect if site.specie.name == new_species]
        )
        defect_new_species_idx = np.array(
            [defect.index(site) for site in defect if site.specie.name == new_species]
        )

        if bulk_new_species_coords.size > 0:  # intrinsic substitution
            # find coords of new species in defect structure, taking into account periodic
            # boundaries
            distance_matrix = np.linalg.norm(
                pbc_diff(bulk_new_species_coords[:, None], defect_new_species_coords),
                axis=2,
            )
            site_matches = distance_matrix.argmin(axis=1)

            if len(np.unique(site_matches)) != len(site_matches):
                raise RuntimeError(
                    "Could not uniquely determine site of new species in defect structure"
                )

            defect_site_idx = list(
                set(np.arange(len(defect_new_species_coords), dtype=int))
                - set(site_matches)
            )[0]

        else:  # extrinsic substitution
            defect_site_idx = 0

        defect_coords = defect_new_species_coords[defect_site_idx]

        # Get the site index of the defect that was used in the VASP calculation
        defect_site_idx = defect_new_species_idx[defect_site_idx]

        # now find the closest old_species site in the bulk structure to the defect site
        # again, make sure to use periodic boundaries
        bulk_old_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == old_species]
        )
        distances = np.linalg.norm(
            pbc_diff(bulk_old_species_coords, defect_coords), axis=1
        )
        original_site_idx = distances.argmin()

        # if there are any other matches with a distance within unique_tolerance of the located
        # site then unique matching failed
        if (
            len(distances[distances < distances[original_site_idx] * unique_tolerance])
            > 1
        ):
            raise RuntimeError(
                "Could not uniquely determine site of old species in bulk structure"
            )

        # currently, original_site_idx is indexed with respect to the old species only.
        # Need to get the index in the full structure
        bulk_coords = np.array([s.frac_coords for s in bulk])
        bulk_site_idx = np.linalg.norm(
            pbc_diff(bulk_coords, bulk_old_species_coords[original_site_idx]), axis=1
        ).argmin()

        # create unrelaxed defect structure
        unrelaxed_defect_structure = bulk.copy()
        unrelaxed_defect_structure.remove_sites([bulk_site_idx])
        # Place defect in same location as output from DFT
        unrelaxed_defect_structure.insert(
            defect_site_idx, new_species, bulk_coords[bulk_site_idx]
        )

    elif defect_type == "vacancy":
        old_species = list(composition_diff.keys())[0]

        bulk_old_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == old_species]
        )
        defect_old_species_coords = np.array(
            [site.frac_coords for site in defect if site.specie.name == old_species]
        )

        # make sure to do take into account periodic boundaries
        distance_matrix = np.linalg.norm(
            pbc_diff(bulk_old_species_coords[:, None], defect_old_species_coords),
            axis=2,
        )
        site_matches = distance_matrix.argmin(axis=0)

        if len(np.unique(site_matches)) != len(site_matches):
            raise RuntimeError(
                "Could not uniquely determine site of vacancy in defect structure"
            )

        original_site_idx = list(
            set(np.arange(len(bulk_old_species_coords), dtype=int)) - set(site_matches)
        )[0]

        # currently, original_site_idx is indexed with respect to the old species only.
        # Need to get the index in the full structure
        bulk_coords = np.array([s.frac_coords for s in bulk])
        bulk_site_idx = np.linalg.norm(
            pbc_diff(bulk_coords, bulk_old_species_coords[original_site_idx]), axis=1
        ).argmin()

        # create unrelaxed defect structure
        unrelaxed_defect_structure = bulk.copy()
        unrelaxed_defect_structure.remove_sites([bulk_site_idx])
        defect_site_idx = None

    elif defect_type == "interstitial":
        new_species = list(composition_diff.keys())[0]

        bulk_new_species_coords = np.array(
            [site.frac_coords for site in bulk if site.specie.name == new_species]
        )
        defect_new_species_coords = np.array(
            [site.frac_coords for site in defect if site.specie.name == new_species]
        )
        defect_new_species_idx = np.array(
            [defect.index(site) for site in defect if site.specie.name == new_species]
        )

        if bulk_new_species_coords.size > 0:  # intrinsic interstitial
            # make sure to take into account periodic boundaries
            distance_matrix = np.linalg.norm(
                pbc_diff(bulk_new_species_coords[:, None], defect_new_species_coords),
                axis=2,
            )
            site_matches = distance_matrix.argmin(axis=1)

            if len(np.unique(site_matches)) != len(site_matches):
                raise RuntimeError(
                    "Could not uniquely determine site of interstitial in defect structure"
                )

            defect_site_idx = list(
                set(np.arange(len(defect_new_species_coords), dtype=int))
                - set(site_matches)
            )[0]

        else:  # extrinsic interstitial
            defect_site_idx = 0

        defect_site_coords = defect_new_species_coords[defect_site_idx]

        # Get the site index of the defect that was used in the VASP calculation
        defect_site_idx = defect_new_species_idx[defect_site_idx]

        # create unrelaxed defect structure
        unrelaxed_defect_structure = bulk.copy()
        # Place defect in same location as output from DFT
        unrelaxed_defect_structure.insert(
            defect_site_idx, new_species, defect_site_coords
        )
        bulk_site_idx = None

    else:
        raise ValueError(f"Invalid defect type: {defect_type}")

    return (
        bulk_site_idx,
        defect_site_idx,
        unrelaxed_defect_structure,
    )


def get_site_mapping_indices(
    structure_a: Structure, structure_b: Structure, threshold=2.0
):
    """
    Reset the position of a partially relaxed structure to its unrelaxed positions.
    The template structure may have a different species ordering to the `input_structure`.
    """

    ## Generate a site matching table between the input and the template
    input_fcoords = [site.frac_coords for site in structure_a]
    template_fcoords = [site.frac_coords for site in structure_b]

    dmat = structure_a.lattice.get_all_distances(input_fcoords, template_fcoords)
    min_dist_with_index = []
    for index in range(len(input_fcoords)):
        dists = dmat[index]
        template_index = dists.argmin()
        current_dist = dists.min()
        min_dist_with_index.append(
            [
                current_dist,
                index,
                template_index,
            ]
        )

        if current_dist > threshold:
            sitea = structure_a[index]
            siteb = structure_b[template_index]
            warnings.warn(
                f"Large site displacement {current_dist:.4f} detected when matching atomic sites:"
                f" {sitea}-> {siteb}."
            )
    return min_dist_with_index


def reorder_unrelaxed_structure(
    unrelaxed_structure: Structure, initial_relax_structure: Structure, threshold=2.0
):
    """
    Reset the position of a partially relaxed structure to its unrelaxed positions.
    The template structure may have a different species ordering to the `input_structure`.
    """

    # Obtain site mapping between the initial_relax_structure and the unrelaxed structure
    mapping = get_site_mapping_indices(
        initial_relax_structure, unrelaxed_structure, threshold=threshold
    )

    # Reorder the unrelaxed_structure so it matches the ordering of the initial_relax_structure (
    # from the actual calculation)
    reordered_sites = [unrelaxed_structure[tmp[2]] for tmp in mapping]
    new_structure = Structure.from_sites(reordered_sites)

    assert len(new_structure) == len(unrelaxed_structure)

    return new_structure


class SingleDefectParser:
    _delocalization_warning_printed = False  # class variable
    # ensures the verbose delocalization analysis warning is only printed once. Needs to be done
    # this way because the current workflow is to create a `SingleDefectParser` object for each
    # defect, and then warning originates from the `run_compatibility()` method of different
    # `SingleDefectParser` instances, so warnings detects each instance as a different source and
    # prints the warning multiple times. When we move to a single function call for all defects
    # (as described above), this can be removed.

    def __init__(
        self,
        defect_entry,
        compatibility=DefectCompatibility(
            plnr_avg_var_tol=0.01,
            plnr_avg_minmax_tol=0.3,
            atomic_site_var_tol=0.025,
            atomic_site_minmax_tol=0.3,
            tot_relax_tol=5.0,
            defect_tot_relax_tol=5.0,
            use_bandfilling=False,  # don't include bandfilling by default
            use_bandedgeshift=False,  # don't include band edge shift by default
        ),
        defect_vr=None,
        bulk_vr=None,
    ):
        """
        Parse a defect object using features that resemble that of a standard
        DefectBuilder object (emmet), but without the requirement of atomate.
        Also allows for use of DefectCompatibility object within pymatgen

        :param defect_entry (DefectEntry): DefectEntry of interest (using the bulk supercell as
        bulk_structure)
            NOTE: to make use of methods within the class, bulk_path and and defect_path
            must exist within the defect_entry parameters class.
        :param compatibility (DefectCompatibility): Compatibility class instance for
            performing charge correction compatibility analysis on defect entry.
        :param defect_vr (Vasprun):
        :param bulk_vr (Vasprun):

        """
        self.defect_entry = defect_entry
        self.compatibility = compatibility
        self.defect_vr = defect_vr
        self.bulk_vr = bulk_vr

    @classmethod
    def from_paths(
        self,
        defect_path,
        bulk_path,
        dielectric,
        charge=None,
        initial_defect_structure=None,
        skip_corrections=False,
        bulk_bandgap_path=None,
        **kwargs,
    ):
        """
        Parse the defect calculation outputs in `defect_path` and return the parsed `DefectEntry`
        object.

        Args:
        defect_path (str): path to defect folder of interest (with vasprun.xml(.gz))
        bulk_path (str): path to bulk folder of interest (with vasprun.xml(.gz))
        dielectric (float or int or 3x1 matrix or 3x3 matrix):
            ionic + static contributions to dielectric constant
        charge (int): charge of defect. If not provided, will be automatically determined
            from the defect calculation outputs (requires POTCARs to be set up with `pymatgen`).
        initial_defect_structure (str):  Path to the unrelaxed defect structure,
            if structure matching with the relaxed defect structure(s) fails.
        skip_corrections (bool): Whether to skip the calculation and application of finite-size
            charge corrections to the defect energy.
        bulk_bandgap_path (str): Path to bulk OUTCAR file for determining the band gap. If the
            VBM/CBM occur at reciprocal space points not included in the bulk supercell calculation,
            you should use this tag to point to a bulk bandstructure calculation instead. If None,
            will use self.defect_entry.parameters["bulk_path"].

        Return:
            Parsed `DefectEntry` object.
        """
        from doped.analysis import defect_entry_from_paths

        kwargs.update({"return SingleDefectParser": True})
        return defect_entry_from_paths(
            defect_path,
            bulk_path,
            dielectric,
            charge=charge,
            initial_defect_structure=initial_defect_structure,
            skip_corrections=skip_corrections,
            bulk_bandgap_path=bulk_bandgap_path,
            **kwargs,
        )

    def freysoldt_loader(self, bulk_locpot=None):
        """
        Load metadata required for performing Freysoldt correction
        requires "bulk_path" and "defect_path" to be loaded to DefectEntry parameters dict.
        Can read gunzipped "LOCPOT.gz" files as well.

        Args:
            bulk_locpot (Locpot): Add bulk Locpot object for expedited parsing.
                If None, will load from file path variable bulk_path
        Return:
            bulk_locpot object for reuse by another defect entry (for expedited parsing)
        """
        if not self.defect_entry.charge:
            # dont need to load locpots if charge is zero
            return None

        if not bulk_locpot:
            bulk_locpot_path, multiple = _get_output_files_and_check_if_multiple(
                "LOCPOT", self.defect_entry.parameters["bulk_path"]
            )
            if multiple:
                warnings.warn(
                    f"Multiple `LOCPOT` files found in bulk directory: "
                    f"{self.defect_entry.parameters['bulk_path']}. Using {bulk_locpot_path} to "
                    f"parse the electrostatic potential and compute the Freysoldt (FNV) charge "
                    f"correction."
                )
            bulk_locpot = get_locpot(bulk_locpot_path)

        defect_locpot_path, multiple = _get_output_files_and_check_if_multiple(
            "LOCPOT", self.defect_entry.parameters["defect_path"]
        )
        if multiple:
            warnings.warn(
                f"Multiple `LOCPOT` files found in defect directory: "
                f"{self.defect_entry.parameters['defect_path']}. Using {defect_locpot_path} to "
                f"parse the electrostatic potential and compute the Freysoldt (FNV) charge "
                f"correction."
            )
        defect_locpot = get_locpot(defect_locpot_path)

        axis_grid = [defect_locpot.get_axis_grid(i) for i in range(3)]
        bulk_planar_averages = [bulk_locpot.get_average_along_axis(i) for i in range(3)]
        defect_planar_averages = [
            defect_locpot.get_average_along_axis(i) for i in range(3)
        ]

        self.defect_entry.parameters.update(
            {
                "axis_grid": axis_grid,
                "bulk_planar_averages": bulk_planar_averages,
                "defect_planar_averages": defect_planar_averages,
                "defect_frac_sc_coords": self.defect_entry.site.frac_coords,
            }
        )
        if "unrelaxed_defect_structure" in self.defect_entry.parameters:
            self.defect_entry.parameters.update(
                {
                    "initial_defect_structure": self.defect_entry.parameters[
                        "unrelaxed_defect_structure"
                    ],
                }
            )

    def kumagai_loader(self, bulk_outcar=None):
        """
        Load metadata required for performing Kumagai correction
        requires "bulk_path" and "defect_path" to be loaded to DefectEntry parameters dict.

        Args:
            bulk_outcar (Outcar): Add bulk Outcar object for expedited parsing.
                If None, will load from file path variable bulk_path
        Return:
            bulk_outcar object for reuse by another defect entry (for expedited parsing)
        """
        if not self.defect_entry.charge:
            # dont need to load outcars if charge is zero
            return None

        if not bulk_outcar:
            bulk_outcar_path, multiple = _get_output_files_and_check_if_multiple(
                "OUTCAR", self.defect_entry.parameters["bulk_path"]
            )
            if multiple:
                warnings.warn(
                    f"Multiple `OUTCAR` files found in bulk directory: "
                    f"{self.defect_entry.parameters['bulk_path']}. Using {bulk_outcar_path} to "
                    f"parse core levels and compute the Kumagai (eFNV) image charge correction."
                )
            bulk_outcar = get_outcar(bulk_outcar_path)
        else:
            bulk_outcar_path = bulk_outcar.filename

        defect_outcar_path, multiple = _get_output_files_and_check_if_multiple(
            "OUTCAR", self.defect_entry.parameters["defect_path"]
        )
        if multiple:
            warnings.warn(
                f"Multiple `OUTCAR` files found in defect directory: "
                f"{self.defect_entry.parameters['defect_path']}. Using {defect_outcar_path} to "
                f"parse core levels and compute the Kumagai (eFNV) image charge correction."
            )
        defect_outcar = get_outcar(defect_outcar_path)

        bulk_atomic_site_averages = bulk_outcar.electrostatic_potential
        defect_atomic_site_averages = defect_outcar.electrostatic_potential
        if not bulk_atomic_site_averages:
            raise ValueError(
                f"Unable to parse atomic core potentials from bulk `OUTCAR` at "
                f"{bulk_outcar_path}. This can happen if `ICORELEVEL` was not set to 0 (= "
                f"default) in the `INCAR`, or if the calculation was finished prematurely with a "
                f"`STOPCAR`. The Kumagai charge correction cannot be computed without this data!"
            )

        if not defect_atomic_site_averages:
            raise ValueError(
                f"Unable to parse atomic core potentials from defect `OUTCAR` at "
                f"{defect_outcar_path}. This can happen if `ICORELEVEL` was not set to 0 (= "
                f"default) in the `INCAR`, or if the calculation was finished prematurely with a "
                f"`STOPCAR`. The Kumagai charge correction cannot be computed without this data!"
            )

        bulk_structure = self.defect_entry.bulk_structure
        bulksites = [site.frac_coords for site in bulk_structure]

        defect_structure = self.defect_entry.parameters["initial_defect_structure"]
        initsites = [site.frac_coords for site in defect_structure]

        distmatrix = bulk_structure.lattice.get_all_distances(
            bulksites, initsites
        )  # first index of this list is bulk index
        min_dist_with_index = [
            [
                min(distmatrix[bulk_index]),
                int(bulk_index),
                int(distmatrix[bulk_index].argmin()),
            ]
            for bulk_index in range(len(distmatrix))
        ]  # list of [min dist, bulk ind, defect ind]

        site_matching_indices = []
        if isinstance(self.defect_entry.defect, (Vacancy, Interstitial)):
            for mindist, bulk_index, defect_index in min_dist_with_index:
                if mindist < 0.5:
                    site_matching_indices.append([bulk_index, defect_index])

        elif isinstance(self.defect_entry.defect, Substitution):
            for mindist, bulk_index, defect_index in min_dist_with_index:
                species_match = (
                    bulk_structure[bulk_index].specie
                    == defect_structure[defect_index].specie
                )
                if mindist < 0.5 and species_match:
                    site_matching_indices.append([bulk_index, defect_index])

        self.defect_entry.parameters.update(
            {
                "bulk_atomic_site_averages": bulk_atomic_site_averages,
                "defect_atomic_site_averages": defect_atomic_site_averages,
                "site_matching_indices": site_matching_indices,
                "defect_frac_sc_coords": self.defect_entry.site.frac_coords,
            }
        )

    def get_stdrd_metadata(self):
        if not self.bulk_vr:
            bulk_vr_path, multiple = _get_output_files_and_check_if_multiple(
                "vasprun.xml", self.defect_entry.parameters["bulk_path"]
            )
            if multiple:
                warnings.warn(
                    f"Multiple `vasprun.xml` files found in bulk directory: "
                    f"{self.defect_entry.parameters['bulk_path']}. Using {bulk_vr_path} to "
                    f"parse the calculation energy and metadata."
                )
            self.bulk_vr = get_vasprun(bulk_vr_path)

        if not self.defect_vr:
            defect_vr_path, multiple = _get_output_files_and_check_if_multiple(
                "vasprun.xml", self.defect_entry.parameters["defect_path"]
            )
            if multiple:
                warnings.warn(
                    f"Multiple `vasprun.xml` files found in defect directory: "
                    f"{self.defect_entry.parameters['defect_path']}. Using {defect_vr_path} to "
                    f"parse the calculation energy and metadata."
                )
            self.defect_vr = get_vasprun(defect_vr_path)

        # standard bulk metadata
        bulk_energy = self.bulk_vr.final_energy
        bulk_sc_structure = self.bulk_vr.initial_structure
        self.defect_entry.parameters.update(
            {"bulk_energy": bulk_energy, "bulk_sc_structure": bulk_sc_structure}
        )

        # standard run metadata
        run_metadata = {}
        run_metadata.update(
            {
                "defect_incar": self.defect_vr.incar,
                "bulk_incar": self.bulk_vr.incar,
                "defect_kpoints": self.defect_vr.kpoints,
                "bulk_kpoints": self.bulk_vr.kpoints,
            }
        )
        run_metadata.update(
            {
                "incar_calctype_summary": {
                    k: self.defect_vr.incar.get(k, None)
                    if self.defect_vr.incar.get(k) not in ["None", "False", False]
                    else None
                    for k in [
                        "LHFCALC",
                        "HFSCREEN",
                        "IVDW",
                        "LUSE_VDW",
                        "LDAU",
                        "METAGGA",
                    ]
                }
            }
        )
        run_metadata.update(
            {
                "potcar_summary": {
                    "pot_spec": [
                        potelt["titel"] for potelt in self.defect_vr.potcar_spec
                    ],
                    "pot_labels": self.defect_vr.potcar_spec,
                    "pot_type": self.defect_vr.run_type,
                }
            }
        )

        self.defect_entry.parameters.update({"run_metadata": run_metadata.copy()})

        # standard defect run metadata
        self.defect_entry.parameters.update(
            {
                "final_defect_structure": self.defect_vr.final_structure,
                "defect_energy": self.defect_vr.final_energy,
            }
        )

        # grab defect energy and eigenvalue information for band filling and localization analysis
        eigenvalues = {
            spincls.value: eigdict.copy()
            for spincls, eigdict in self.defect_vr.eigenvalues.items()
        }
        kpoint_weights = self.defect_vr.actual_kpoints_weights[:]
        self.defect_entry.parameters.update(
            {"eigenvalues": eigenvalues, "kpoint_weights": kpoint_weights}
        )

    def get_bulk_gap_data(
        self, bulk_bandgap_path=None, use_MP=False, mpid=None, api_key=None
    ):
        """Get bulk gap data from bulk OUTCAR file, or OUTCAR located at `actual_bulk_path`.

        Alternatively, one can specify query the Materials Project (MP) database for the bulk gap
        data, using `use_MP = True`, in which case the MP entry with the lowest number ID and
        composition matching the bulk will be used, or  the MP ID (mpid) of the bulk material to
        use can be specified. This is not recommended as it will correspond to a
        severely-underestimated GGA DFT bandgap!

        Args:
            bulk_bandgap_path (str): Path to bulk OUTCAR file for determining the band gap. If
                the VBM/CBM occur at reciprocal space points not included in the bulk supercell
                calculation, you should use this tag to point to a bulk bandstructure calculation
                instead. If None, will use self.defect_entry.parameters["bulk_path"].
            use_MP (bool): If True, will query the Materials Project database for the bulk gap
                data.
            mpid (str): If provided, will query the Materials Project database for the bulk gap
                data, using this Materials Project ID.
            api_key (str): Materials API key to access database.
        """
        if not self.bulk_vr:
            bulk_vr_path, multiple = _get_output_files_and_check_if_multiple(
                "vasprun.xml", self.defect_entry.parameters["bulk_path"]
            )
            if multiple:
                warnings.warn(
                    f"Multiple `vasprun.xml` files found in bulk directory: "
                    f"{self.defect_entry.parameters['bulk_path']}. Using {bulk_vr_path} to "
                    f"parse the calculation energy and metadata."
                )
            self.bulk_vr = get_vasprun(bulk_vr_path)

        bulk_sc_structure = self.bulk_vr.initial_structure

        vbm, cbm, bandgap = None, None, None
        gap_parameters = {}

        if use_MP and mpid is None:
            try:
                with MPRester(api_key=api_key) as mp:
                    tmp_mplist = mp.get_entries_in_chemsys(
                        list(bulk_sc_structure.symbol_set)
                    )
                mplist = [
                    ment.entry_id
                    for ment in tmp_mplist
                    if ment.composition.reduced_composition
                    == bulk_sc_structure.composition.reduced_composition
                ]
            except Exception as exc:
                raise ValueError(
                    f"Error with querying MPRester for"
                    f" {bulk_sc_structure.composition.reduced_formula}:"
                ) from exc

            mpid_fit_list = []
            for trial_mpid in mplist:
                with MPRester(api_key=api_key) as mp:
                    mpstruct = mp.get_structure_by_material_id(trial_mpid)
                if StructureMatcher(
                    primitive_cell=True,
                    scale=False,
                    attempt_supercell=True,
                    allow_subset=False,
                ).fit(bulk_sc_structure, mpstruct):
                    mpid_fit_list.append(trial_mpid)

            if len(mpid_fit_list) == 1:
                mpid = mpid_fit_list[0]
                print(f"Single mp-id found for bulk structure:{mpid}.")
            elif len(mpid_fit_list) > 1:
                num_mpid_list = [int(mp.split("" - "")[1]) for mp in mpid_fit_list]
                num_mpid_list.sort()
                mpid = "mp-" + str(num_mpid_list[0])
                print(
                    f"Multiple mp-ids found for bulk structure:{mpid_fit_list}. Will use lowest "
                    f"number mpid for bulk band structure = {mpid}."
                )
            else:
                print(
                    "Could not find bulk structure in MP database after tying the following "
                    f"list:\n{mplist}"
                )
                mpid = None

        if mpid is not None:
            print(f"Using user-provided mp-id for bulk structure: {mpid}.")
            with MPRester(api_key=api_key) as mp:
                bs = mp.get_bandstructure_by_material_id(mpid)
            if bs:
                cbm = bs.get_cbm()["energy"]
                vbm = bs.get_vbm()["energy"]
                bandgap = bs.get_band_gap()["energy"]
                gap_parameters.update(
                    {"MP_gga_BScalc_data": bs.get_band_gap().copy()}
                )  # contains gap kpt transition

        if vbm is None or bandgap is None or cbm is None or not bulk_bandgap_path:
            if mpid and bandgap is None:
                print(
                    f"WARNING: Mpid {mpid} was provided, but no bandstructure entry currently "
                    "exists for it. \nReverting to use of bulk supercell calculation for band "
                    "edge extrema."
                )

            gap_parameters.update(
                {"MP_gga_BScalc_data": None}
            )  # to signal no MP BS is used
            bandgap, cbm, vbm, _ = self.bulk_vr.eigenvalue_band_properties

        if bulk_bandgap_path:
            print(f"Using actual bulk path: {bulk_bandgap_path}")
            actual_bulk_vr_path, multiple = _get_output_files_and_check_if_multiple(
                "vasprun.xml", bulk_bandgap_path
            )
            if multiple:
                warnings.warn(
                    f"Multiple `vasprun.xml` files found in specified directory: "
                    f"{bulk_bandgap_path}. Using {actual_bulk_vr_path} to  parse the calculation "
                    f"energy and metadata."
                )
            actual_bulk_vr = get_vasprun(actual_bulk_vr_path)
            bandgap, cbm, vbm, _ = actual_bulk_vr.eigenvalue_band_properties

        gap_parameters.update({"mpid": mpid, "cbm": cbm, "vbm": vbm, "gap": bandgap})
        self.defect_entry.parameters.update(gap_parameters)

    def run_compatibility(self):
        # Set potalign so pymatgen can calculate bandfilling for 'neutral' defects
        # (possible for resonant dopants etc.)
        if (
            self.defect_entry.charge == 0
            and "potalign" not in self.defect_entry.parameters
        ):
            self.defect_entry.parameters["potalign"] = 0

        self.defect_entry = self.compatibility.process_entry(self.defect_entry)

        if "delocalization_meta" in self.defect_entry.parameters:
            delocalization_meta = self.defect_entry.parameters["delocalization_meta"]
            if (
                "plnr_avg" in delocalization_meta
                and not delocalization_meta["plnr_avg"]["is_compatible"]
            ) or (
                "atomic_site" in delocalization_meta
                and not delocalization_meta["atomic_site"]["is_compatible"]
            ):
                specific_delocalized_warning = (
                    f"Delocalization analysis has indicated that {self.defect_entry.name} with "
                    f"charge {self.defect_entry.charge:+} may not be compatible with the chosen "
                    f"charge correction."
                )
                general_delocalization_warning = """
Note: Defects throwing a "delocalization analysis" warning may require a larger supercell for
accurate total energies. Recommended to look at the correction plots (i.e. run 
`get_correction_freysoldt(DefectEntry,...,plot=True)` from
`doped.corrections`) to visually determine if the charge 
correction scheme is still appropriate (replace 'freysoldt' with 'kumagai' if using anisotropic 
correction). You can also change the DefectCompatibility() tolerance settings via the 
`compatibility` parameter in `SingleDefectParser.from_paths()`."""
                warnings.warn(message=specific_delocalized_warning)
                if not self._delocalization_warning_printed:
                    warnings.warn(
                        message=general_delocalization_warning
                    )  # should only print once
                    SingleDefectParser._delocalization_warning_printed = (
                        True  # don't print again
                    )

        if "num_hole_vbm" in self.defect_entry.parameters:
            if (
                self.compatibility.free_chg_cutoff
                < self.defect_entry.parameters["num_hole_vbm"]
            ) or (
                self.compatibility.free_chg_cutoff
                < self.defect_entry.parameters["num_elec_cbm"]
            ):
                num_holes = self.defect_entry.parameters["num_hole_vbm"]
                num_electrons = self.defect_entry.parameters["num_elec_cbm"]
                warnings.warn(
                    f"Eigenvalue analysis has determined that `num_hole_vbm` (= {num_holes}) or "
                    f"`num_elec_cbm` (= {num_electrons}) is significant (>2.1) for "
                    f"{self.defect_entry.name} with charge {self.defect_entry.charge}:+, "
                    f"indicating that there are many free charges in this defect supercell "
                    f"calculation and so the defect charge correction is unlikely to be accurate."
                )
                if "freysoldt_meta" in self.defect_entry.parameters:
                    frey_meta = self.defect_entry.parameters["freysoldt_meta"]
                    frey_corr = (
                        frey_meta["freysoldt_electrostatic"]
                        + frey_meta["freysoldt_potential_alignment_correction"]
                    )
                    self.defect_entry.corrections.update(
                        {"charge_correction": frey_corr}
                    )
                elif "kumagai_meta" in self.defect_entry.parameters:
                    kumagai_meta = self.defect_entry.parameters["kumagai_meta"]
                    kumagai_corr = (
                        kumagai_meta["kumagai_electrostatic"]
                        + kumagai_meta["kumagai_potential_alignment_correction"]
                    )
                    self.defect_entry.corrections.update(
                        {"charge_correction": kumagai_corr}
                    )

        if (
            self.defect_entry.charge != 0
            and self.defect_entry.corrections.get("charge_correction", None) is None
        ):
            warnings.warn(
                f"No charge correction computed for {self.defect_entry.name} with "
                f"charge {self.defect_entry.charge:+}, indicating problems with the "
                f"required data for the charge correction (i.e. dielectric constant, "
                f"LOCPOT files for Freysoldt correction, OUTCAR (with ICORELEVEL = 0) "
                f"for Kumagai correction etc)."
            )


class PostProcess:
    def __init__(self, root_fldr, mpid=None, mapi_key=None):
        """
        Post processing object for charged point-defect calculations.

        Args:
            root_fldr (str): path (relative) to directory
                in which data of charged point-defect calculations for
                a particular system are to be found;
            mpid (str): Materials Project ID of bulk structure;
                format "mp-X", where X is an integer;
            mapi_key (str): Materials API key to access database.

        """
        self._root_fldr = root_fldr
        self._mpid = mpid
        self._mapi_key = mapi_key
        self._substitution_species = set()

    def parse_defect_calculations(self):
        """
        Parses the defect calculations as DefectEntry objects,
        from a PyCDT root_fldr file structure.
        Charge correction is missing in the first run.
        """
        parsed_defects = []
        subfolders = glob.glob(os.path.join(self._root_fldr, "vac_*"))
        subfolders += glob.glob(os.path.join(self._root_fldr, "as_*"))
        subfolders += glob.glob(os.path.join(self._root_fldr, "sub_*"))
        subfolders += glob.glob(os.path.join(self._root_fldr, "inter_*"))

        def get_vr_and_check_locpot(fldr):
            vr_file = os.path.join(fldr, "vasprun.xml")
            if not (os.path.exists(vr_file) or os.path.exists(vr_file + ".gz")):
                warnings.warn("{} doesn't exit".format(vr_file))
                error_msg = ": Failure, vasprun.xml doesn't exist."
                return (None, error_msg)  # Further processing is not useful

            try:
                vr = get_vasprun(vr_file, parse_potcar_file=False)
            except:
                warnings.warn("Couldn't parse {}".format(vr_file))
                error_msg = ": Failure, couldn't parse vasprun.xml file."
                return (None, error_msg)

            if not vr.converged:
                warnings.warn("Vasp calculation at {} not converged".format(fldr))
                error_msg = ": Failure, Vasp calculation not converged."
                return (None, error_msg)  # Further processing is not useful

            # Check if locpot exists
            locpot_file = os.path.join(fldr, "LOCPOT")
            if not (os.path.exists(locpot_file) or os.path.exists(locpot_file + ".gz")):
                warnings.warn("{} doesn't exit".format(locpot_file))
                error_msg = ": Failure, LOCPOT doesn't exist"
                return (None, error_msg)  # Further processing is not useful

            return (vr, None)

        def get_encut_from_potcar(fldr):
            potcar_file = os.path.join(fldr, "POTCAR")
            if not os.path.exists(potcar_file):
                warnings.warn("Not POTCAR in {} to parse ENCUT".format(fldr))
                error_msg = ": Failure, No POTCAR file."
                return (None, error_msg)  # Further processing is not useful

            try:
                potcar = Potcar.from_file(potcar_file)
            except:
                warnings.warn("Couldn't parse {}".format(potcar_file))
                error_msg = ": Failure, couldn't read POTCAR file."
                return (None, error_msg)

            encut = max(ptcr_sngl.enmax for ptcr_sngl in potcar)
            return (encut, None)

        # get bulk entry information first
        fldr = os.path.join(self._root_fldr, "bulk")
        vr, error_msg = get_vr_and_check_locpot(fldr)
        if error_msg:
            raise ValueError(f"Abandoning parsing of the calculations: {error_msg}")
        bulk_energy = vr.final_energy
        bulk_sc_struct = vr.final_structure
        try:
            encut = vr.incar["ENCUT"]
        except:  # ENCUT not specified in INCAR. Read from POTCAR
            encut, error_msg = get_encut_from_potcar(fldr)
            if error_msg:
                raise ValueError(f"Abandoning parsing of the calculations: {error_msg}")

        trans_dict = loadfn(os.path.join(fldr, "transformation.json"), cls=MontyDecoder)
        supercell_size = trans_dict["supercell"]

        bulk_file_path = fldr
        bulk_entry = ComputedStructureEntry(
            bulk_sc_struct,
            bulk_energy,
            data={
                "bulk_path": bulk_file_path,
                "encut": encut,
                "supercell_size": supercell_size,
            },
        )

        # get defect entry information
        for fldr in subfolders:
            fldr_name = os.path.split(fldr)[1]
            chrg_fldrs = glob.glob(os.path.join(fldr, "charge*"))
            for chrg_fldr in chrg_fldrs:
                trans_dict = loadfn(
                    os.path.join(chrg_fldr, "transformation.json"), cls=MontyDecoder
                )
                chrg = trans_dict["charge"]
                vr, error_msg = get_vr_and_check_locpot(chrg_fldr)
                if error_msg:
                    warnings.warn("Parsing the rest of the calculations")
                    continue
                if (
                    "substitution_specie" in trans_dict
                    and trans_dict["substitution_specie"]
                    not in bulk_sc_struct.symbol_set
                ):
                    self._substitution_species.add(trans_dict["substitution_specie"])
                elif (
                    "inter" in trans_dict["defect_type"]
                    and trans_dict["defect_site"].specie.symbol
                    not in bulk_sc_struct.symbol_set
                ):
                    # added because extrinsic interstitials don't have
                    # "substitution_specie" character...
                    trans_dict["substitution_specie"] = trans_dict[
                        "defect_site"
                    ].specie.symbol
                    self._substitution_species.add(
                        trans_dict["defect_site"].specie.symbol
                    )

                defect_type = trans_dict.get("defect_type", None)
                energy = vr.final_energy
                try:
                    encut = vr.incar["ENCUT"]
                except:  # ENCUT not specified in INCAR. Read from POTCAR
                    encut, error_msg = get_encut_from_potcar(chrg_fldr)
                    if error_msg:
                        warnings.warn(
                            "Not able to determine ENCUT " "in {}".format(fldr_name)
                        )
                        warnings.warn("Parsing the rest of the " "calculations")
                        continue

                comp_data = {
                    "bulk_path": bulk_file_path,
                    "defect_path": chrg_fldr,
                    "encut": encut,
                    "fldr_name": fldr_name,
                    "supercell_size": supercell_size,
                }
                if "substitution_specie" in trans_dict:
                    comp_data["substitution_specie"] = trans_dict["substitution_specie"]

                # create Defect Object as dict, then load to DefectEntry object
                defect_dict = {
                    "structure": bulk_sc_struct,
                    "charge": chrg,
                    "@module": "pymatgen.analysis.defects.core",
                }
                defect_site = trans_dict["defect_supercell_site"]
                if "vac_" in defect_type:
                    defect_dict["@class"] = "Vacancy"
                elif "as_" in defect_type or "sub_" in defect_type:
                    defect_dict["@class"] = "Substitution"
                    substitution_specie = trans_dict["substitution_specie"]
                    defect_site = PeriodicSite(
                        substitution_specie,
                        defect_site.frac_coords,
                        defect_site.lattice,
                        coords_are_cartesian=False,
                    )
                elif "int_" in defect_type:
                    defect_dict["@class"] = "Interstitial"
                else:
                    raise ValueError(
                        "defect type {} not recognized...".format(defect_type)
                    )

                defect_dict.update({"defect_site": defect_site})
                defect = MontyDecoder().process_decoded(defect_dict)
                parsed_defects.append(
                    DefectEntry(defect, energy - bulk_energy, parameters=comp_data)
                )

        try:
            parsed_defects_data = {}
            parsed_defects_data["bulk_entry"] = bulk_entry
            parsed_defects_data["defects"] = parsed_defects
            return parsed_defects_data
        except:
            return {}  # Return Null dict due to failure

    def get_vbm_bandgap(self):
        """
        Returns the valence band maximum (float) of the structure with
        MP-ID mpid.

        Args:
            mpid (str): MP-ID for which the valence band maximum is to
                be fetched from the Materials Project database
        """
        vbm, bandgap = None, None

        if self._mpid is not None:
            with MPRester(api_key=self._mapi_key) as mp:
                bs = mp.get_bandstructure_by_material_id(self._mpid)
            if bs:
                vbm = bs.get_vbm()["energy"]
                bandgap = bs.get_band_gap()["energy"]

        if vbm is None or bandgap is None:
            if self._mpid:
                warnings.warn(
                    "Mpid {} was provided, but no bandstructure entry currently exists for it. "
                    "Reverting to use of bulk calculation.".format(self._mpid)
                )
            else:
                warnings.warn(
                    "No mp-id provided, will fetch CBM/VBM details from the "
                    "bulk calculation."
                )
            warnings.warn(
                "This may not be appropriate if the VBM/CBM occur at reciprocal points "
                "not included in the bulk calculation."
            )
            vr = get_vasprun(
                os.path.join(self._root_fldr, "bulk", "vasprun.xml"),
                parse_potcar_file=False,
            )
            bandgap = vr.eigenvalue_band_properties[0]
            vbm = vr.eigenvalue_band_properties[2]

        return (vbm, bandgap)

    def get_chempot_limits(self):
        """
        Returns atomic chempots from bulk_composition based on data in
        the materials project database. This is abstractly handled in the
        ChemPotAnalyzer

        Note to user: If personal phase diagram desired,
            option exists in the pycdt.core.chemical_potentials to setup,
            run and parse personal phase diagrams for purposes of chemical potentials
        """

        if self._mpid:
            cpa = _chemical_potentials.MPChemPotAnalyzer(
                mpid=self._mpid,
                sub_species=self._substitution_species,
                mapi_key=self._mapi_key,
            )
        else:
            bulkvr = get_vasprun(
                os.path.join(self._root_fldr, "bulk", "vasprun.xml"),
                parse_potcar_file=False,
            )
            if not bulkvr:
                msg = "Could not fetch computed entry for atomic chempots!"
                warnings.warn(msg)
                raise ValueError(msg)
            cpa = _chemical_potentials.MPChemPotAnalyzer(
                bulk_ce=bulkvr.get_computed_entry(),
                sub_species=self._substitution_species,
                mapi_key=self._mapi_key,
            )

        chem_lims = cpa.analyze_GGA_chempots()

        return chem_lims

    def parse_dielectric_calculation(self):
        """
        Parses the "vasprun.xml(.gz)" file in subdirectory "dielectric" of
        root directory root_fldr and returns the average of the trace
        of the dielectric tensor.

        Args:
            root_fldr (str):
                root directory where subdirectory "dielectric" is expected
        Returns:
            eps (float):
                average of the trace of the dielectric tensor
        """

        try:
            vr = get_vasprun(
                os.path.join(self._root_fldr, "dielectric", "vasprun.xml"),
                parse_potcar_file=False,
            )
        except:
            warnings.warn("Parsing Dielectric calculation failed")
            return None

        eps_ion = vr.epsilon_ionic
        eps_stat = vr.epsilon_static

        eps = []
        for i in range(len(eps_ion)):
            eps.append([e[0] + e[1] for e in zip(eps_ion[i], eps_stat[i])])

        return eps

    def compile_all(self):
        """
        Run to get all post processing objects as dictionary

        note: still need to implement
            1) ability for substitutional atomic chempots
            2) incorporated charge corrections for defects
        """
        output = self.parse_defect_calculations()
        output["epsilon"] = self.parse_dielectric_calculation()
        output["mu_range"] = self.get_chempot_limits()
        vbm, gap = self.get_vbm_bandgap()
        output["vbm"] = vbm
        output["gap"] = gap

        return output



import os
import tarfile
from shutil import copyfile
import numpy as np

from monty.tempfile import ScratchDir
from pymatgen.io.vasp import Locpot
from pymatgen.util.testing import PymatgenTest
from pymatgen.analysis.defects.core import DefectEntry, Vacancy
from pymatgen.core.sites import PeriodicSite

from doped.corrections import (
    freysoldt_correction_from_paths,
    kumagai_correction_from_paths,
    get_correction_freysoldt,
    get_correction_kumagai,
)

test_files_dir = os.path.join(os.path.dirname(__file__), "../doped/pycdt/test_files")


class FilePathCorrectionsTest(PymatgenTest):
    def test_freysoldt_and_kumagai(self):
        # create scratch directory with files....
        # having to do it all at once to minimize amount of time copying over to Scratch Directory
        with ScratchDir("."):
            # setup with fake Locpot object copied over
            copyfile(
                os.path.join(test_files_dir, "test_path_files.tar.gz"),
                "./test_path_files.tar.gz",
            )
            tar = tarfile.open("test_path_files.tar.gz")
            tar.extractall()
            tar.close()
            blocpot = Locpot.from_file(os.path.join(test_files_dir, "bLOCPOT.gz"))
            blocpot.write_file("test_path_files/bulk/LOCPOT")
            dlocpot = Locpot.from_file(os.path.join(test_files_dir, "dLOCPOT.gz"))
            dlocpot.write_file("test_path_files/sub_1_Sb_on_Ga/charge_2/LOCPOT")

            fcc = freysoldt_correction_from_paths(
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
                "test_path_files/bulk/",
                18.12,
                2,
                plot=True,
                filename="test_freysoldt_correction",
            )
            self.assertAlmostEqual(
                fcc, -1.4954476868106865
            )  # note this has been updated from the
            # pycdt version, because there they used a `transformation.json` that gave an
            # incorrect `initial_defect_structure` (corresponding to primitive rather than bulk)
            self.assertTrue(os.path.exists("test_freysoldt_correction_axis1.pdf"))

            kcc = kumagai_correction_from_paths(
                "test_path_files/sub_1_Sb_on_Ga/charge_2/",
                "test_path_files/bulk/",
                18.12,
                2,
                plot=True,
                filename="test_kumagai_correction",
            )
            self.assertAlmostEqual(kcc, 0.638776853061614)
            self.assertTrue(os.path.exists("test_kumagai_correction.pdf"))


class FiniteSizeChargeCorrectionTest(PymatgenTest):
    """
    Test functions for getting freysoldt and kumagai corrections
    """

    def setUp(self):
        self.dielectric = 15.0

        struc = PymatgenTest.get_structure("VO2")
        struc.make_supercell(3)
        vac = Vacancy(struc, struc.sites[0], charge=-3)

        # load neccessary parameters for defect_entry to make use
        # of Freysoldt and Kumagai corrections
        p = {}
        ids = vac.generate_defect_structure(1)
        abc = struc.lattice.abc
        axisdata = [np.arange(0.0, lattval, 0.2) for lattval in abc]
        bldata = [
            np.array([1.0 for u in np.arange(0.0, lattval, 0.2)]) for lattval in abc
        ]
        dldata = [
            np.array(
                [
                    (-1 - np.cos(2 * np.pi * u / lattval))
                    for u in np.arange(0.0, lattval, 0.2)
                ]
            )
            for lattval in abc
        ]
        p.update(
            {
                "axis_grid": axisdata,
                "bulk_planar_averages": bldata,
                "defect_planar_averages": dldata,
                "initial_defect_structure": ids,
                "defect_frac_sc_coords": struc.sites[0].frac_coords,
                "bulk_sc_structure": struc,
            }
        )

        bulk_atomic_site_averages, defect_atomic_site_averages = [], []
        defect_site_with_sc_lattice = PeriodicSite(
            struc.sites[0].specie,
            struc.sites[0].coords,
            struc.lattice,
            coords_are_cartesian=True,
        )
        max_dist = 9.6
        pert_amnt = 1.0
        for site_ind, site in enumerate(struc.sites):
            if site.specie.symbol == "O":
                Oval = -30.6825
                bulk_atomic_site_averages.append(Oval)
                if site_ind:
                    dist_to_defect = site.distance_and_image(
                        defect_site_with_sc_lattice
                    )[0]
                    defect_site_val = (
                        Oval
                        - 0.3
                        + pert_amnt * ((max_dist - dist_to_defect) / max_dist) ** 2
                    )
                    defect_atomic_site_averages.append(defect_site_val)
            else:
                Vval = -51.6833
                bulk_atomic_site_averages.append(Vval)
                if site_ind:
                    dist_to_defect = site.distance_and_image(
                        defect_site_with_sc_lattice
                    )[0]
                    defect_site_val = (
                        Vval
                        - 0.3
                        + pert_amnt * ((max_dist - dist_to_defect) / max_dist) ** 2
                    )
                    defect_atomic_site_averages.append(defect_site_val)

        site_matching_indices = [
            [ind, ind - 1] for ind in range(len(struc.sites)) if ind != 0
        ]

        p.update(
            {
                "bulk_atomic_site_averages": bulk_atomic_site_averages,
                "defect_atomic_site_averages": defect_atomic_site_averages,
                "site_matching_indices": site_matching_indices,
            }
        )
        self.defect_entry = DefectEntry(vac, 0.0, parameters=p)

    def test_get_correction_freysoldt(self):
        freyout = get_correction_freysoldt(
            self.defect_entry, self.dielectric, partflag="All", axis=None
        )
        self.assertEqual(freyout, 5.445950368792991)

        freyout = get_correction_freysoldt(
            self.defect_entry, self.dielectric, partflag="AllSplit", axis=None
        )
        self.assertAlmostEqual(freyout[0], 0.975893)
        self.assertAlmostEqual(freyout[1], 4.4700573687929905)
        self.assertAlmostEqual(freyout[2], 5.445950368792991)

    def test_get_correction_kumagai(self):
        kumagaiout = get_correction_kumagai(
            self.defect_entry, self.dielectric, partflag="AllSplit"
        )
        self.assertAlmostEqual(kumagaiout[0], 0.9763991294314076)
        self.assertAlmostEqual(kumagaiout[1], 0.2579750033409367)
        self.assertAlmostEqual(kumagaiout[2], 1.2343741327723443)
